/*
 *	Sherlock HTML/ASCII Chewer
 *
 *	(c) 1997--1999 Martin Mares, <mj@atrey.karlin.mff.cuni.cz>
 */

/*
 *	Calling convention:
 *
 *	Command line: <doc-url> <base-url>
 *
 *	fd[0] = data file
 *	fd[1] = put chewed-out information to
 *	fd[2] = log file
 *
 *	Exit code: 0=distilled, 1=no data written
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <sys/time.h>

#include <lib/lib.h>
#include <lib/string.h>
#include <lib/url.h>
#include <lib/searchdb.h>
#include <charset/unicode.h>

#include "htmlchew.h"

/* Configuration parameters */

static int trace = 0;
static int comment_mode = 0;
static int show_errors = 0;
static uns freq_threshold = 1000000;
static int quote_hack = 0;
static int freq_limits = 0;

static struct cfitem cfitems[] = {
  { "Trace", CI_INT, &trace },
  { "ShowErrors", CI_INT, &show_errors },
  { "CommentMode", CI_INT, &comment_mode },
  { "FreqThreshold", CI_INT, &freq_threshold },
  { "FreqLimits", CI_INT, &freq_limits },
  { "QuoteHack", CI_INT, &quote_hack },
  { NULL, CI_STOP, NULL }
};

/* Global variables */

static int is_html;
static struct url base_url, doc_url;
static byte urlbuf1[MAX_URL_SIZE], urlbuf2[MAX_URL_SIZE];
static byte urlbuf3[MAX_URL_SIZE], urlbuf4[MAX_URL_SIZE];
static byte *orig_base_url, *orig_doc_url;
static byte *meta_description, *meta_title;

#define RC_INDEX 1			/* META ROBOTS and similar switches */
#define RC_FOLLOW 2

static uns robot_control = RC_INDEX | RC_FOLLOW;

/* Reading of input characters and decoding of UTF8 */

#define UNGET_STK_SIZE 256

static char unget_stack[UNGET_STK_SIZE];
static int unget_sp = 0;

static int
get_char(void)
{
  int c;

  if (unget_sp)
    return unget_stack[--unget_sp];
  for(;;)
    {
      c = getchar();
      if (c < 0)				/* EOF */
	return c;
      if (c < 0x80)				/* ASCII character */
	return c;
      else if (IS_UTF8(c))
	{
	  uns u;
	  if (c >= 0xf0)			/* Out of 16-bit code range */
	    {
	      while ((c = getchar()) >= 0x80 && c < 0xc0)
		;
	      u = UNI_REPLACEMENT;
	    }
	  else if (c >= 0xe0)		/* 16-bit code */
	    {
	      u = c & 0x0f;
	      c = getchar();
	      u = (u << 6) | (c & 0x3f);
	      c = getchar();
	      u = (u << 6) | (c & 0x3f);
	    }
	  else				/* 11-bit code */
	    {
	      u = c & 0x1f;
	      c = getchar();
	      u = (u << 6) | (c & 0x3f);
	    }
	  return u;
	}
      else if (trace >= 20)
	printf("..Invalid character %02x\n", c);
    }
}

static void
unget_char(int i)
{
  if (unget_sp < UNGET_STK_SIZE)
    unget_stack[unget_sp++] = i;
}

/* Parsing and decoding of HTML entities, returns UniCode value of the entity */

static int
decode_entity(byte *e)
{
  const struct entity *n;
  struct entity ent;

  if (e[0] == '#')
    {
      n = &ent;
      if (e[1] == 'x' || e[1] == 'X')
	ent.value = strtol(e+2, NULL, 16);
      else
	ent.value = atol(e+1);
    }
  else n = is_entity(e, strlen(e));
  if (n)
    {
      if (trace >= 50)
	printf("..&%s; = %02x\n", e, n->value);
      return n->value;
    }
  else if (trace >= 50)
    printf("..&%s; ???\n", e);
  return 0;
}

static int
get_entity(void)
{
  byte buf[80];
  uns i;
  int c;

  i = 0;
  c = get_char();
  if (c == '#')
    {
      buf[i++] = c;
      c = get_char();
    }
  while (c >= 0 && c <= 0x100 && Calnum(c))
    {
      if (i >= sizeof(buf) - 1)		/* Suspiciously long entity, try to recover */
	return 0;
      buf[i++] = c;
      c = get_char();
    }
  if (c < 0)
    return 0;
  buf[i] = 0;
  if (Cspace(c))			/* Ends with white space */
    {
      unget_char(c);
      return decode_entity(buf);
    }
  else if (c == ';')			/* Standard ending */
    return decode_entity(buf);
  unget_char(c);			/* Syntax error */
  while (i > 0)
    unget_char(buf[--i]);
  return '&';
}

/* Parse document URL and base URL */

static void
prep_base(byte *d, byte *b)
{
  int err;

  orig_base_url = b;
  orig_doc_url = d;
  if (   (err = url_canon_split(b, urlbuf1, urlbuf2, &base_url))
	 || (err = url_canon_split(d, urlbuf3, urlbuf4, &doc_url)))
    {
      printf("-00%02d\n", err);
      exit(1);
    }
}

/*
 * Test word for validity.
 * Rule: Reject everything shorter than 3 chars with exception of 2-letter words
 */

static int
is_valid_word(word *w)
{
  if (Ualpha(w[0]) && Ualpha(w[1]))
    return 1;
  return Ustrlen(w) >= 3;
}

/* Document examples */

struct docex {
  byte *first;
  uns pos, max, nospace;
  char *name;
};

static byte title_ex[80], body_ex[256];

static struct docex title_docex = { title_ex, 0, sizeof(title_ex)-1, 1, "TITLE" };
static struct docex body_docex = { body_ex, 0, sizeof(body_ex)-1, 1, "BODY" };
static struct docex *current_ex = NULL;

static void
ex_set(struct docex *x)
{
  if (trace >= 10)
    printf("..Context = %s\n", x ? x->name : "NULL");
  if (x)
    {
      if (!x->nospace)
	{
	  if (x->pos && x->pos < x->max)
	    x->first[x->pos++] = ' ';
	  x->nospace = 1;
	}
      if (x->pos >= x->max)
	current_ex = NULL;
      else
	current_ex = x;
    }
  else
    current_ex = x;
}

static void
add_ex_char(word c)
{
  struct docex *x = current_ex;

  if (!x)				/* Shouldn't happen */
    return;
  if (Uspace(c) || !Uprint(c))
    {
      x->nospace = 1;
      return;
    }
  if (x->nospace)
    {
      if (x->pos && x->pos < x->max)
	x->first[x->pos++] = ' ';
      x->nospace = 0;
    }
  if (x->pos <= x->max - UTF8_SPACE(c))
    {
      byte *z = x->first + x->pos;
      PUT_UTF8(z,c);
      x->pos = z - x->first;
    }
  else
    current_ex = NULL;
}

static void
flush_ex(byte f, struct docex *x)
{
  if (x->pos >= x->max)
    {
      while (x->pos > 0 && Calnum(x->first[x->pos - 1]))
	x->pos--;
      if (x->pos > 0 && Cblank(x->first[x->pos - 1]))
	x->pos--;
    }
  if (!x->pos)
    return;

  x->first[x->pos] = 0;
  printf("%c%s\n", f, x->first);
}

/* Word gathering */

#define WW_HASH 256
#define HTML_MAX_WORD_LEN 64

struct ww {
  struct ww *next;
  uns count;
  byte classes;
  byte w[1];
};

static struct ww *ww_hash[WW_HASH];

#define QC_BUGGY 0
#define QC_STRONG 1
#define QC_LINK 2
#define QC_DEF 3
#define QC_H_SMALL 4
#define QC_H_BIG 5
#define QC_TITLE 6
#define QC_KEYWORD 7
#define QC_MAX 8

static int class_bonus[QC_MAX] = { -8, 1, 1, 2, 3, 4, 8, 12 };
static uns class_counter[QC_MAX];
static uns class_limit[QC_MAX] = { ~0, ~0, ~0, ~0, ~0, ~0, 30, 10 };

static void
wattr_classify(struct ww *w, uns wa)
{
  uns c = 0;
  int i;

  if (wa & A_KEYWORD)
    c |= 1 << QC_KEYWORD;
  if (wa & A_TITLE)
    c |= 1 << QC_TITLE;
  if (wa & (A_H1 | A_H2 | A_H3))
    c |= 1 << QC_H_BIG;
  if (wa & (A_H4 | A_H5 | A_H6))
    c |= 1 << QC_H_SMALL;
  if (wa & A_DFN)
    c |= 1 << QC_DEF;
  if (wa & A_ANCHOR)
    c |= 1 << QC_LINK;
  if (wa & (A_EM | A_STRONG))
    c |= 1 << QC_STRONG;
  if (! (wa & A_BODY))			/* Out of body expirience :) */
    c |= 1 << QC_BUGGY;

  for(i=0; i<QC_MAX; i++)
    if (c & (1 << i))
      class_counter[i]++;

  w->classes |= c;
}

static int
classify_word(struct ww *w)
{
  int q = 0;
  uns c = w->classes;
  int i;

  /* Apply per-class limits if requested */
  if (freq_limits)
    {
      for(i=0; i<QC_MAX; i++)
	if ((c & (1 << i)) && class_counter[i] >= class_limit[i])
	  c &= ~(1 << i);
    }

  /* Find highest class and use its bonus */
  for(i=QC_MAX-1; i>=0; i--)
    if (c & (1 << i))
      break;
  if (i >= 0)
    q += class_bonus[i];

  /* Apply frequency threshold if needed */
  if (w->count >= freq_threshold)
    q++;

  return q;
}

static void
add_word(word *wo, uns wattrs)
{
  struct ww *w;
  uns hash, len;
  byte *b;
  byte xbuf[HTML_MAX_WORD_LEN*4];

#if 0
  if (trace >= 100)
    printf("..[%s] %x %d\n", word, wattrs, q);
#endif

  if (! (robot_control & RC_INDEX))
    return;

  hash = len = ucs2_to_utf8(xbuf, wo);
  for(b = xbuf; *b; b++)
    hash = hash*37U + *b;
  hash &= (WW_HASH - 1);

  for(w = ww_hash[hash]; w; w = w->next)
    if (!strcmp(xbuf, w->w))
      break;

  if (w)
    w->count++;
  else
    {
      w = xmalloc(sizeof(struct ww) + len);
      w->next = ww_hash[hash];
      ww_hash[hash] = w;
      w->classes = 0;
      w->count = 1;
      strcpy(w->w, xbuf);
    }
  wattr_classify(w, wattrs);
}

static void
flush_words(void)
{
  uns cnt = 0;
  struct ww *w;
  uns h, len;
  int q;

  for(h = 0; h < WW_HASH; h++)
    for(w = ww_hash[h]; w; w = w->next)
      {
	len = strlen(w->w);
	if (cnt + len > 80)
	  {
	    cnt = 0;
	    putchar('\n');
	  }
	putchar(cnt ? ' ' : 'W');
	cnt += len + 1;
	q = classify_word(w) + WQ_OFFSET;
	if (q < 0)
	  q = 0;
	if (q > WQ_MASK)
	  q = WQ_MASK;
	if (q != WQ_OFFSET)
	  putchar(0x80 + q);
	fputs(w->w, stdout);
      }
  if (cnt)
    putchar('\n');
}

/* Word buffering */

static word wordbuf[HTML_MAX_WORD_LEN];
static uns wordlen;
static uns attrs, wattrs;

static void
flush_word(void)
{
  if (!wordlen)
    return;
  if (wordlen < HTML_MAX_WORD_LEN)
    {
      wattrs &= ~A_IS_WS;
      wordbuf[wordlen] = 0;
      if (is_valid_word(wordbuf))
	add_word(wordbuf, wattrs);
    }
  wordlen = 0;
  wattrs = 0;
}

static inline void
add_char(uns c)
{
  if (attrs & (A_STYLE | A_SCRIPT))	/* Doomed to ignorance */
    return;
  if (Ualnum(c))
    {
      if (wordlen < HTML_MAX_WORD_LEN)
	{
	  wordbuf[wordlen++] = Utoupper(c);
	  wattrs |= attrs;
	}
    }
  else if (wordlen)			/* Redundant condition, speedup only */
    flush_word();
  if (current_ex)
    add_ex_char(c);
}

static void
add_string(byte *s)
{
  while (*s)
    {
      uns u;
      if (IS_UTF8(*s))
	GET_UTF8(s, u);
      else
	u = *s++;
      add_char(u);
    }
}

/* References */

#define REF_A 0
#define REF_IMG 1
#define REF_APPLET 2
#define REF_CLASS 3
#define REF_FORM 4

#define REF_HASH 64

struct ref {
  struct ref *next;
  byte type;
  byte ref[1];
};

static struct ref *ref_hash[REF_HASH];

static void
add_ref(byte *ref, uns type)
{
  byte buf1[MAX_URL_SIZE], buf2[MAX_URL_SIZE], buf3[MAX_URL_SIZE];
  int err;
  struct url url;
  uns hash, len;
  byte *k;
  struct ref *r;

  if (! (robot_control & RC_FOLLOW))	/* Following of references explicitly forbidden */
    return;

  if (type != REF_CLASS)
    {
      if (err = url_deescape(ref, buf1))
	goto bad;
      if (err = url_split(buf1, &url, buf2))
	goto bad;
      if (err = url_normalize(&url, &base_url))
	goto bad;
      if (err = url_canonicalize(&url))
	goto bad;
      if (   !strcmp(url.protocol, doc_url.protocol)
	     && !strcmp(url.host, doc_url.host)
	     && url.port == doc_url.port
	     && !strcmp(url.rest, doc_url.rest))
	{
	  if (trace >= 20)
	    puts("..SELF-REF");
	  return;
	}
      if (err = url_pack(&url, buf3))
	goto bad;
      if (err = url_enescape(buf3, buf2))
	goto bad;
      ref = buf2;
    }

  if (trace >= 10)
    printf("..REF %s %d\n", ref, type);

  hash = len = strlen(ref);
  for(k = ref; *k; k++)
    hash = (hash * 37U) + *k;
  hash &= (REF_HASH - 1);
  for(r = ref_hash[hash]; r; r = r->next)
    if (!strcmp(r->ref, ref) && r->type == type)
      break;

  if (!r)
    {
      r = xmalloc(sizeof(struct ref) + len);
      r->next = ref_hash[hash];
      ref_hash[hash] = r;
      r->type = type;
      strcpy(r->ref, ref);
      printf("%c%s\n", "RIJjf"[type], ref);
    }
  return;

bad:
  if (show_errors)
    log(L_WARN "%s: invalid reference to %s [%d] (00%02d %s)", orig_doc_url,
	ref, type, err, url_error(err));
}

/* Process META tag */

static int
chew_meta_robots(byte *t)
{
  uns r = RC_FOLLOW | RC_INDEX;
  byte *p;
  byte z;

  if (!strcasecmp(t, "ALL"))
    return r;
  if (!strcasecmp(t, "NONE"))
    return 0;
  while (*t)
    {
      if (Cspace(*t) || *t == ',')
	{
	  t++;
	  continue;
	}
      p = t;
      while (*t && *t != ',' && !Cspace(*t))
	t++;
      z = *t;
      *t = 0;
      if (!strcasecmp(p, "NOINDEX"))
	r &= ~RC_INDEX;
      else if (!strcasecmp(p, "NOFOLLOW"))
	r &= ~RC_FOLLOW;
      else if (!strcasecmp(p, "INDEX"))
	r |= RC_INDEX;
      else if (!strcasecmp(p, "FOLLOW"))
	r |= RC_FOLLOW;
      else if (*p)
	return -1;
      *t = z;
    }
  return r;
}

static void
chew_meta(byte *name, byte *text)
{
  if (trace > 30)
    log(L_DEBUG "META %s=%s", name, text);
  if (!strcasecmp(name, "description"))
    meta_description = stralloc(text);
  else if (!strcasecmp(name, "robots"))
    {
      int r = chew_meta_robots(text);
      if (r < 0)
	{
	err:
	  log(L_WARN "%s: Unknown META %s=%s", orig_doc_url, name, text);
	  return;
	}
      printf(".Meta robot control: %sINDEX, %sFOLLOW\n",
	     (r & RC_INDEX) ? "" : "NO",
	     (r & RC_FOLLOW) ? "" : "NO");
      robot_control = r;
    }
  else if (!strcasecmp(name, "keywords"))
    {
      uns old_attrs;
      struct docex *old_ex;

      flush_word();
      old_attrs = attrs;
      attrs = A_KEYWORD | A_BODY;
      old_ex = current_ex;
      current_ex = NULL;
      add_string(text);
      flush_word();
      attrs = old_attrs;
      current_ex = old_ex;
    }
  else if (!strcasecmp(name, "title"))
    meta_title = stralloc(text);
  else if (!strcasecmp(name, "refresh"))
    {
      byte *k = strchr(text, ';');
      if (k)
	{
	  k++;
	  while (Cspace(*k))
	    k++;
	  if (strncasecmp(k, "URL=", 4))
	    goto err;
	  add_ref(k+4, REF_A);
	}
      else if (k = strchr(text, ','))
	{
	  k++;
	  while (Cspace(*k))
	    k++;
	  if (*k)
	    add_ref(k, REF_A);
	}
    }
#ifdef WARN_UNKNOWN_METAS
  else if (strcasecmp(name, "author")	/* Useless META's */
	   && strcasecmp(name, "generator")
	   && strcasecmp(name, "formatter"))
    goto err;
#endif
}

/* Chew an already parsed HTML tag */

static byte **attr_names, **attr_vals;
static uns attr_count;

static byte *
getarg(byte *name)
{
  uns i;

  for(i=0; i<attr_count; i++)
    if (!strcmp(attr_names[i], name))
      return attr_vals[i] ? attr_vals[i] : (byte *) "";
  return NULL;
}

static void
chew_tag(byte *name, byte **anames, byte **avals, uns acount)
{
  const struct tag *t;
  byte *k, *l;

  attr_names = anames;
  attr_vals = avals;
  attr_count = acount;
  t = is_tag(name, strlen(name));
  if (trace >= 50)
    {
      uns i;
      printf("..%s", name);
      for(i=0; i<acount; i++)
	printf(",%s=%s", anames[i], avals[i]);
      if (t)
	printf(" -> %d,%x", t->type, t->arg);
      putchar('\n');
    }
  if (t)
    {
      if (t->type != T_ON && t->type != T_OFF) /* New word... */
	flush_word();
      switch (t->type)
	{
	case T_WSEP:			/* New word only */
	  break;
	case T_ON:			/* Switch an attribute on */
	  if (t->arg & A_IS_WS)
	    flush_word();
	  attrs |= t->arg;
	  break;
	case T_OFF:			/* Switch an attribute off */
	  if (t->arg & A_IS_WS)
	    flush_word();
	  attrs &= ~t->arg;
	  break;
	case T_BASE:			/* Set base URL */
	  if (k = getarg("HREF"))
	    {
	      if (url_canon_split(k, urlbuf1, urlbuf2, &base_url))
		{
		  log(L_WARN "%s: invalid base URL %s", orig_doc_url, k);
		  printf(".Invalid base URL: %s\n", k);
		  prep_base(orig_doc_url, orig_base_url); /* Use the original one */
		}
	    }
	  break;
	case T_META:			/* Meta-information */
	  k = getarg("NAME");
	  if (!k)
	    k = getarg("HTTP-EQUIV");
	  l = getarg("CONTENT");
	  if (k && l)
	    chew_meta(k, l);
	  else
	    log(L_WARN "%s: suspicious META tag", orig_doc_url);
	  break;
	case T_A:			/* Anchor */
	case T_LINK:			/* <LINK>s usually have HREF and can be parsed as <A>s */
	  if (t->arg)			/* End */
	    attrs &= ~A_ANCHOR;
	  else
	    {
	      if (getarg("NAME"))	/* Internal label */
		break;
	      if (k = getarg("HREF"))
		{
		  add_ref(k, REF_A);
		  attrs |= A_ANCHOR;
		}
	    }
	  break;
	case T_IMG:			/* An image */
	  if (k = getarg("ALT"))
	    add_string(k);
	  if (k = getarg("SRC"))
	    add_ref(k, REF_IMG);
	  break;
	case T_FRAME:			/* Frame -- an extremely ugly Netscape extension :( */
	  if (k = getarg("NAME"))
	    add_string(k);
	  if (k =getarg("SRC"))
	    add_ref(k, REF_A);
	  break;
	case T_APPLET:			/* An applet */
	  if (k = getarg("ALT"))
	    add_string(k);
	  if (k = getarg("CODEBASE"))
	    add_ref(k, REF_APPLET);
	  if (k = getarg("CODE"))
	    add_ref(k, REF_CLASS);
	  break;
	case T_FORM:			/* A form */
	  if (k = getarg("ACTION"))
	    add_ref(k, REF_FORM);
	  break;
	case T_INPUT:			/* A form line */
	  if (k = getarg("NAME"))
	    add_string(k);
	  break;
	case T_MAP:			/* An image map */
	  if (k = getarg("NAME"))
	    add_string(k);
	  break;
	case T_TITLE:			/* Document title */
	  if (t->arg)
	    {
	      attrs &= ~A_TITLE;
	      ex_set(NULL);
	    }
	  else
	    {
	      attrs |= A_TITLE;
	      ex_set(&title_docex);
	    }
	  break;
	case T_BODY:			/* Document body */
	  if (t->arg)
	    {
	      attrs &= ~A_BODY;
	      ex_set(NULL);
	    }
	  else
	    {
	      attrs |= A_BODY;
	      ex_set(&body_docex);
	    }
	  break;
	case T_ISINDEX:			/* Search allowed */
	  puts(".ISINDEX");
	  break;
	case T_BLOCKQUOTE:
	  if (k = getarg("CITE"))
	    add_ref(k, REF_A);
	  break;
	default:
	  die("Unknown tag type");
	}
    }
}

/* Parse a HTML meta-tag (<!...>) */

static void
parse_meta_tag(void)
{
  int c;

  if (trace >= 50)
    puts("..Meta-tag");
  c = get_char();
  for(;;)
    {
      if (c < 0 || c == '>')
	return;
      if (c == '-')			/* Maybe a comment */
	{
	  c = get_char();
	  if (c != '-')
	    continue;
	  switch (comment_mode)
	    {
	    case 0:			/* Standard behaviour */
	      c = get_char();
	      for(;;)
		{
		  if (c < 0)
		    return;
		  if (c == '-')
		    {
		      c = get_char();
		      if (c == '-')
			break;
		    }
		  else
		    c = get_char();
		}
	      break;
	    case 1:			/* Silly implementation: end with ">" */
	      for(;;)
		{
		  c = get_char();
		  if (c < 0 || c == '>')
		    return;
		}
	      /* Cannot get here */
	    default:			/* Buggy Netscape compatibility: end with "-->" */
	      c = get_char();
	      for(;;)
		{
		  if (c < 0)
		    return;
		  if (c == '-')
		    {
		      c = get_char();
		      while (c == '-')
			{
			  c = get_char();
			  if (c == '>' || c < 0)
			    return;
			}
		    }
		  else
		    c = get_char();
		}
	    }
	}
      else if (c == '"' || c == '\'')	/* Quoted part */
	{
	  int k = c;
	  int hacking_quotes = 0;
	  while ((c = get_char()) != k)
	    {
	      if (c < 0)
		return;
	      if (c == '\r' || c == '\n')
		{
		  if (quote_hack == 1)
		    return;
		  if (quote_hack)
		    hacking_quotes = 1;
		}
	      if (hacking_quotes && c == '>')
		return;
	    }
	}
      c = get_char();
    }
}

/* Parse a HTML tag */

#define MAX_ATTRS 16

static void
parse_tag(void)
{					/* Ugly state-machine-like code :-| */
  byte buf[1024];
  byte *aname[MAX_ATTRS+1], *aval[MAX_ATTRS+1];
  int c, attr;
  byte *k, *stop;

  stop = buf + sizeof(buf) - 8;		/* Parse tag name */
  k = buf;
  attr = 0;
  for(;;)
    {
      c = get_char();
      if (c < 0)
	return;
      if (c == '>')
	goto finished;
      if (Uspace(c) || c == '\r' || c == '\n')
	break;
      if (c >= 0x80)			/* Tag names: non-ASCII characters are replaced by DEL */
	c = 0x7f;
      if (k < stop)
	*k++ = Cupcase(c);
    }
  if (k < stop)
    *k++ = 0;

  for(;;)				/* Parse attributes */
    {
      while (c >= 0 && (Uspace(c) || c == '\r' || c == '\n'))
	c = get_char();
      if (c < 0)
	return;
      if (c == '>')
	break;
      aname[attr] = k;
      for(;;)
	{
	  if (c < 0)
	    return;
	  if (c == '>')
	    {
	      aval[attr++] = NULL;
	      goto finished;
	    }
	  if (Uspace(c) || c == '\r' || c == '\n')
	    {
	      aval[attr] = NULL;
	      goto cont;
	    }
	  if (c == '=')
	    {
	      c = get_char();
	      break;
	    }
	  if (c >= 0x80)		/* Attribute names: non-ASCII characters are replaced by DEL */
	    c = 0x7f;
	  if (k < stop)
	    *k++ = Cupcase(c);
	  c = get_char();
	}
      if (k < stop)
	*k++ = 0;
      aval[attr] = k;
      for(;;)
	{
	  if (c == '>')
	    {
	      attr++;
	      goto finished;
	    }
	  if (Uspace(c) || c == '\r' || c == '\n')
	    break;
	  if (c == '"' || c == '\'')
	    {
	      int d = c;

	      c = get_char();
	      while (c != d)
		{
		  int hacking_quotes = 0;
		  if (c < 0)
		    return;
		  if (c == '\r' || c == '\n')
		    {
		      if (quote_hack == 1)
			return;
		      if (quote_hack)
			hacking_quotes = 1;
		    }
		  if (hacking_quotes && c == '>')
		    return;
		  if (c == '&')		/* Honor entities in attribute values */
		    {
		      c = get_entity();
		      if (!c)
			continue;
		    }
		  if (k < stop)		/* Attribute values: non-ASCII characters are encoded in UTF-8 */
		    PUT_UTF8(k, c);
		  c = get_char();
		}
	    }
	  else if (k < stop)
	    PUT_UTF8(k, c);
	  c = get_char();
	  if (c < 0)
	    return;
	}
    cont:
      if (k < stop)
	*k++ = 0;
      if (attr < MAX_ATTRS)
	attr++;
    }

finished:
  if (k < stop)
    {
      *k = 0;
      chew_tag(buf, aname, aval, attr);
    }
}

/* Main chewing (gum | function) for HTML documents */

static void
chew_html(void)
{
  int c;

  while ((c = get_char()) >= 0)
    {
      if (c == '<')			/* Tag or meta-tag */
	{
	  c = get_char();
	  if (c < 0)
	    break;
	  if (c == '!')
	    parse_meta_tag();
	  else
	    {
	      unget_char(c);
	      parse_tag();
	    }
	  continue;
	}
      if (c == '&')			/* Entity */
	{
	  c = get_entity();
	  if (!c)
	    continue;
	}
      add_char(c);
    }
  flush_word();
}

/* The same for text documents */

static void
chew_text(void)
{
  int c;

  attrs = A_BODY;			/* Everything is considered document body */
  ex_set(&body_docex);

  while ((c = get_char()) >= 0)
    add_char(c);
  flush_word();
}

/* Flushing of gathered info */

static void
flush_all(void)
{
  if (meta_title)
    printf("H%s\n", meta_title);
  else
    flush_ex('H', &title_docex);
  if (meta_description)
    printf("C%s\n", meta_description);
  else
    flush_ex('C', &body_docex);
  flush_words();
}

/* MAIN */

int
main(int argc, char **argv)
{
  char *k;

  initlog(argv[0]);
  if (argc != 2 && argc != 3)
    die("Invalid arguments. Usage: (html|text)chew <doc-URL> [<base-URL>]");
  k = strrchr(argv[0], '/');
  k = k ? k+1 : argv[0];
  is_html = !strcmp(k, "htmlchew");
  prep_base(argv[1], argv[2] ? argv[2] : argv[1]);
  cf_read("cf/htmlchew", cfitems);
  if (is_html)
    chew_html();
  else
    chew_text();
  flush_all();
  return 0;
}
