/*
 *	Sherlock Gatherer -- Character Set Processing
 *
 *	(c) 1997 Martin Mares, <mj@atrey.karlin.mff.cuni.cz>
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <alloca.h>

#include "gather.h"

#include <lib/string.h>
#include <lib/url.h>
#include <charset/unicode.h>

/* CONFIGURATION */

int trace_charsets;
#define TRACE trace_charsets
#define XTRACE (TRACE > 1)

struct deflang {
  node n;
  byte *host_patt;
  byte *lang_list;
};

struct langmap {
  node n;
  byte *src, *dest;
};

static list def_langs, lang_maps;

#define MAX_CHARSETS 32
#define UTF8_CHARSET 31
static byte *charset_names[MAX_CHARSETS];
static word *charset_to_ucs[MAX_CHARSETS];
static uns num_charsets;
static byte *fallback_charset, *fail_fallback_charset;

#define MAX_LANGS 32
#define DEFAULT_LANG 31
static byte *lang_names[MAX_LANGS];
static u32 lang_charsets[MAX_LANGS];
static byte lang_charset_order[MAX_LANGS][MAX_CHARSETS];
static uns num_langs, believe_explicit_charset, judge_threshold;

struct uni_entry {
  word code;
  u32 usual;
  u32 forbidden;
};

#define HASH_SKIP 1733

static struct uni_entry *uni_hash;
static uns uni_hash_size = 1024;
static uns uni_hash_mask, num_slots, num_coll, max_coll;

static byte *snerr = "Syntax error";

static byte *
parse_deflang(struct cfitem *c, byte *b)
{
  byte *w[2];
  struct deflang *f = xmalloc(sizeof(struct deflang));

  if (wordsplit(b, w, 2) != 2)
    return snerr;
  f->host_patt = stralloc(w[0]);
  f->lang_list = stralloc(w[1]);
  add_tail(&def_langs, &f->n);
  return NULL;
}

static byte *
parse_langmap(struct cfitem *c, byte *b)
{
  byte *w[2];
  struct langmap *f = xmalloc(sizeof(struct langmap));

  if (wordsplit(b, w, 2) != 2)
    return snerr;
  f->src = stralloc(w[0]);
  f->dest = stralloc(w[1]);
  add_tail(&lang_maps, &f->n);
  return NULL;
}

static int
lookup_charset(byte *name)
{
  uns i;

  for(i=0; i<MAX_CHARSETS; i++)
    if (charset_names[i] && !strcasecmp(charset_names[i], name))
      return i;
  return -1;
}

static int
lookup_language(byte *name)
{
  uns i;

  for(i=0; i<MAX_LANGS; i++)
    if (lang_names[i] && !strcasecmp(lang_names[i], name))
      return i;
  return -1;
}

static void
load_charset(word *k,  byte *name)
{
  FILE *f;
  byte buf[256];
  uns code, ucs, lin;

  if (! (f = fopen(name, "r")))
    die("Unable to load charset file %s", name);
  lin = 0;
  while (lin++, fgets(buf, sizeof(buf), f))
    {
      byte *z = strchr(buf, '\n');
      if (!z)
	die("%s: Line %d too long", name, lin);
      *z = 0;
      if (!buf[0] || buf[0] == '#')
	continue;
      if (sscanf(buf, "%x\t%x", &code, &ucs) != 2
	  || code >= 256 || ucs >= 0x10000)
	die("%s: Syntax error on line %d", name, lin);
      k[code] = ucs;
    }
  fclose(f);
}

static byte *
parse_charset(struct cfitem *c, byte *b)
{
  byte *w[2];
  uns i;

  if (wordsplit(b, w, 2) != 2)
    return snerr;
  if (lookup_charset(w[0]) >= 0)
    return "Charset already defined";
  if (!strcasecmp(w[1], "INTERNAL-UTF-8"))
    {
      i = UTF8_CHARSET;
      if (charset_names[i])
	return "UTF-8 already defined";
      die("UTF-8 not supported yet");	/* FIXME */
    }
  else
    {
      if (num_charsets >= MAX_CHARSETS - 1)
	return "Too many charsets";
      i = num_charsets++;
    }
  charset_names[i] = stralloc(w[0]);
  charset_to_ucs[i] = xmalloc(sizeof(word) * 256);
  bzero(charset_to_ucs[i], sizeof(word) * 256);
  if (TRACE)
    log(L_DEBUG "Charset %d is %s (%s)", i, w[0], w[1]);
  if (i != UTF8_CHARSET)
    load_charset(charset_to_ucs[i], w[1]);
  return NULL;
}

static byte *
parse_langset(struct cfitem *c, byte *b)
{
  byte *w[33];
  int i, j, n, lan;
  u32 z;

  if ((n = wordsplit(b, w, 33)) < 2)
    return snerr;
  if (lookup_language(w[0]) >= 0)
    return "Language already defined";
  if (!strcmp(w[0], "*"))
    lan = DEFAULT_LANG;
  else
    {
      if (num_langs >= MAX_LANGS - 1)
	return "Too many languages";
      lan = num_langs++;
    }
  lang_names[lan] = stralloc(w[0]);
  z = 0;
  for(i=1; i<n; i++)
    {
      if ((j = lookup_charset(w[i])) < 0)
	return "Unknown charset";
      z |= (1 << j);
      lang_charset_order[lan][i-1] = j;
    }
  while (i < MAX_CHARSETS + 1)
    lang_charset_order[lan][i++ -1] = ~0;
  lang_charsets[lan] = z;
  if (TRACE)
    log(L_DEBUG "Language %d is %s, charsets %08x", lan, w[0], z);
  return NULL;
}

static void
init_hash(void)
{
  if (!uni_hash)
    {
      uni_hash = xmalloc(uni_hash_size * sizeof(struct uni_entry));
      bzero(uni_hash, uni_hash_size * sizeof(struct uni_entry));
      uni_hash_mask = uni_hash_size - 1;
    }
}

static inline uns
hash(uns code)
{
  return (code * 313) & uni_hash_mask;
}

static inline struct uni_entry *
lookup_entry(uns code)
{
  uns j;

  j = hash(code);
  while (uni_hash[j].code)
    {
      if (uni_hash[j].code == code)
	return &uni_hash[j];
      j = (j + HASH_SKIP) & uni_hash_mask;
    }
  return NULL;
}

static struct uni_entry *
new_entry(uns code)
{
  uns j, c;

  j = hash(code);
  c = 0;
  if (num_slots >= uni_hash_size)
    die("Character code hash table overflow, increase its size");
  while (uni_hash[j].code)
    {
      if (uni_hash[j].code == code)
	return &uni_hash[j];
      c++;
      j = (j + HASH_SKIP) & uni_hash_mask;
    }
  if (c)
    {
      if (c > max_coll)
	max_coll = c;
      num_coll++;
    }
  num_slots++;
  uni_hash[j].code = code;
  return &uni_hash[j];
}

static byte *
parse_charlist(struct cfitem *c, byte *b)
{
  int forbid = (c->name[1] == 'F');
  byte *w[64];
  int i, l, n, lang;
  uns code, lo, hi;
  struct uni_entry *e;

  if ((n = wordsplit(b, w, 64)) < 2)
    return snerr;
  init_hash();
  if ((lang = lookup_language(w[0])) < 0)
    return "Unknown language";
  for(i=1; i<n; i++)
    {
      l = strlen(w[i]);
      if (l != 4 && l != 9)
	return "UCS-2 code/range syntax error";
      if (l == 4)
	{
	  if (sscanf(w[i], "%x", &lo) != 1)
	    return "Invalid UCS-2 character code";
	  hi = lo;
	}
      else
	{
	  if (sscanf(w[i], "%x-%x", &lo, &hi) != 2)
	    return "Invalid UCS-2 code range";
	}
      if (lo > hi)
	return "Invalid code range";
      if (hi >= 0x10000)
	return "UCS-2 Character code out of range";
      if (!lo)
	{
	  if (!forbid)
	    return "Null character is always illegal and cannot be redefined";
	  lo++;
	}
      for(code=lo; code<=hi; code++)
	{
	  u32 m = (1 << lang);
	  e = new_entry(code);
	  if (forbid)
	    {
	      e->forbidden |= m;
	      e->usual &= ~m;
	    }
	  else
	    {
	      e->usual |= m;
	      e->forbidden &= ~m;
	    }
	}
    }
  return NULL;
}

static struct cfitem cfitems[] = {
  { "Trace", CI_INT, &trace_charsets },
  { "DefLang", CI_FUNCTION, parse_deflang },
  { "LangMap", CI_FUNCTION, parse_langmap },
  { "HashSize", CI_INT, &uni_hash_size },
  { "Charset", CI_FUNCTION, parse_charset },
  { "LangSet", CI_FUNCTION, parse_langset },
  { "CForbid", CI_FUNCTION, parse_charlist },
  { "CUsual", CI_FUNCTION, parse_charlist },
  { "Fallback", CI_STRING, &fallback_charset },
  { "Believe", CI_INT, &believe_explicit_charset },
  { "JudgeThreshold", CI_INT, &judge_threshold },
  { "Failback", CI_STRING, &fail_fallback_charset },
  { NULL, CI_STOP, NULL }
};

void
charset_init(void)
{
  init_list(&def_langs);
  init_list(&lang_maps);
  cf_read("cf/charsets", cfitems);
  if (TRACE)
    log(L_DEBUG "Code tables: %d of %d entries, %d collisions (%d max)",
	num_slots, uni_hash_size, num_coll, max_coll);
}

/* MANAGEMENT OF LANGUAGE AND CHARSET SETS */

static byte *
guess_language(struct odes *o, byte *host)
{
  struct deflang *d;

  DO_FOR_ALL(d, def_langs)
    if (match_pattern_nocase(d->host_patt, host))
      {
	if (TRACE)
	  log(L_DEBUG "Using default language `%s' for host %s", d->lang_list, host);
	return d->lang_list;
      }
  die("guess_language: DefLang list doesn't give answer to `%s'", host);
}

static byte *
lookup_langmap(byte *l)
{
  struct langmap *m;

  DO_FOR_ALL(m, lang_maps)
    if (!strcasecmp(m->src, l))
      return m->dest;
  return l;
}

static byte *guessed_fallback;

static u32
guess_langmask(byte *url, byte *lans)
{
  byte *w[32];
  byte *l, *z;
  int n, i, k;
  u32 mask = 0;
  byte *buf = alloca(strlen(lans) + 1);

  strcpy(buf, lans);
  n = wordsplit(buf, w, 32);
  if (n < 0)
    {
      log(L_ERROR "%s: Found more than 32 languages, strange...", url);
      n = 32;
    }
  for(i=0; i<n; i++)
    {
      l = w[i];
      while ((k = lookup_language(lookup_langmap(l))) < 0)
	{
	  if (z = strrchr(l, '-'))	/* Drop off the last suffix */
	    *z = 0;
	  else
	    {
	      k = DEFAULT_LANG;
	      break;
	    }
	}
      mask |= (1 << k);
    }

  if (!mask)
    mask = (1 << DEFAULT_LANG);
  if (TRACE)
    log(L_DEBUG "Language mask is %X", mask);
  return mask;
}

static u32
guess_charmask(u32 lans)
{
  u32 cs = 0;
  int i;

  guessed_fallback = NULL;
  for(i=0; i<MAX_LANGS; i++)
    if (lans & (1 << i))
      {
	cs |= lang_charsets[i];
	if (!guessed_fallback)
	  guessed_fallback = charset_names[lang_charset_order[i][0]];
      }
  return cs;
}

/* CHARSET RECOGNITION (Well, at least an attempt...) */

static uns bad_count[MAX_CHARSETS];
static uns usual_count[MAX_CHARSETS];

static void
reset_stats(void)
{
  bzero(bad_count, sizeof(bad_count));
  bzero(usual_count, sizeof(usual_count));
}

static int				/* 0=no more charsets, 1=certain, -1=uncertain */
charset_judge(u32 *pmask, int force)
{
  u32 mask = *pmask;
  uns i, max, max2, maxi;

  max = 0;
  max2 = 0;
  maxi = 0;
  for(i=0; i<MAX_CHARSETS - 1; i++)
    if (mask & (1 << i))
      {
	if (bad_count[i])
	  mask &= ~(1 << i);
	else
	  {
	    uns uc = usual_count[i];
	    if (uc >= max)
	      {
		max2 = max;
		max = uc;
		maxi = i;
	      }
	    else if (uc > max2)
	      max2 = uc;
	  }
      }

  if (!mask)
    {
      *pmask = 0;
      return 0;
    }

  if (mask == (1U << maxi) || max2 + judge_threshold < max || (force & max2 < max))
    {
      *pmask = (1 << maxi);
      return 1;
    }

  *pmask = mask;
  return -1;
}

static int				/* 0=no more charsets, 1=certain, -1=uncertain */
update_stats(byte *buf, uns len, u32 *pmask, u32 langs)
{
  byte *end = buf + len;
  byte *k;
  word *ct, ucs;
  u32 mask = *pmask;
  uns cs, bc, uc;
  struct uni_entry *e;
  u32 lang2;

  langs &= ~(1 << DEFAULT_LANG);
  lang2 = langs | (1 << DEFAULT_LANG);
  for(cs=0; cs<MAX_CHARSETS - 1; cs++)
    if (mask & (1 << cs))
      {
	bc = bad_count[cs];
	uc = usual_count[cs];
	ct = charset_to_ucs[cs];
	for(k=buf; k<end; k++)
	  {
	    ucs = ct[*k];
	    if (ucs)
	      {
		if (e = lookup_entry(ucs))
		  {
		    if ((e->forbidden & langs) == langs)
		      bc++;
		    else if (e->usual & lang2)
		      uc++;
		    else if (e->forbidden & (1 << DEFAULT_LANG))
		      bc++;
		  }
	      }
	    else
	      bc++;
	  }
	bad_count[cs] = bc;
	usual_count[cs] = uc;
      }

  return charset_judge(pmask, 0);
}

/* MAIN ENTRY POINT */

int
guess_charset(struct odes *o, byte *url, byte *host)
{
  byte *cs = find_aval(o, 'c');
  byte *lan = find_aval(o, 'l');
  u32 lan_mask, char_mask;
  byte *fallback, *best;
  int trying, final;

  if (!lan)				/* Language selection */
    lan = guess_language(o, host);
  else if (TRACE)
    log(L_DEBUG "%s: Got language `%s'", url, lan);
  lan_mask = guess_langmask(url, lan);

  char_mask = guess_charmask(lan_mask);	/* Charset set selection */
  if (cs)
    {
      if ((trying = lookup_charset(cs)) < 0)
	{
	  if ((trying = lookup_charset(fallback_charset)) < 0)
	    die("Fallback charset %s not known", fallback_charset);
	  fallback = fallback_charset;
	}
      else
	fallback = cs;
      char_mask |= (1 << trying);
    }
  else
    {
      if (! (fallback = guessed_fallback))
	fallback = fallback_charset;
      trying = -1;
    }
  if (TRACE)
    log(L_DEBUG "Possible charsets are %08x, fallback=%s, trying=%d", char_mask, fallback, trying);

  if (believe_explicit_charset && trying >= 0)
    {
      if (TRACE)
	log(L_DEBUG "Believing in explicit charset %d", trying);
      final = trying;
    }
  else					/* Do the magic charset vaticination process */
    {
      if (dt_read_head() < 0)
	goto nocharset;
      reset_stats();
      while (dt_in_size)
	{
	  if (update_stats(dt_inbuf, dt_in_size, &char_mask, lan_mask) >= 0) /* Certain or null list */
	    break;
	  if (trying >= 0)		/* We got an explicit charset and we're trying to test it */
	    {
	      if (char_mask & (1 << trying))
		{
		  if (TRACE)
		    log(L_DEBUG "Explicit charset %d is valid", trying);
		  final = trying;
		  goto done;
		}
	      if (TRACE)
		log(L_DEBUG "Explicit charset %d doesn't seem to be valid", trying);
	      trying = -1;
	    }
	  if (dt_read_next() < 0)
	    goto nocharset;
	}
      charset_judge(&char_mask, 1);	/* Threshold comparison disabled */
      if (!char_mask)
	{
	nocharset:
	  if (TRACE)
	    log(L_DEBUG "Charset detection failed, falling back to %s", fallback);
	  if ((final = lookup_charset(fallback)) < 0)
	    die("Fallback charset %s unknown", fallback);
	}
      else if (! (char_mask & (char_mask - 1)))	/* Certain */
	{
	  final = log2(char_mask);
	  if (TRACE)
	    log(L_DEBUG "Guessed charset %d after examining %d bytes", final, dt_where);
	}
      else
	{
	  int i, j, k;
	  for(i=0; i<MAX_LANGS; i++)
	    if (lan_mask & (1 << i))
	      for(j=0; j<MAX_CHARSETS; j++)
		{
		  k = lang_charset_order[i][j];
		  if (k >= MAX_CHARSETS)
		    break;
		  if (char_mask & (1 << k))
		    {
		      if (TRACE)
			log(L_DEBUG "Guessed charsets %X, %d wins", char_mask, k);
		      final = k;
		      goto done;
		    }
		}
	  log(L_ERROR "BUG! Resulting charset and language sets don't intersect (%X,%X)!", char_mask, lan_mask);
	  goto nocharset;
	}
    done:
    }

  best = charset_names[final];
  if (TRACE)
    log(L_DEBUG "%s: Charset=%s", url, best);
  if (cs && strcasecmp(cs, best))
    {
      byte *z = alloca(64);
      sprintf(z, "Orig-Charset: %s", cs);
      add_attr(o, NULL, '.', z);
    }
  set_attr(o, 'c', best);

  return strcasecmp(best, "utf-8");
}

/* CHARSET CONVERSION */

static int cc_charset;
static word *cc_conv;

void
char_convert_init(struct odes *o)
{
  byte *cs = find_aval(o, 'c');

  if (TRACE)
    log(L_DEBUG "Recoding to canonical charset");
  if (!cs || (cc_charset = lookup_charset(cs)) < 0)
    die("char_convert_init: invalid charset");
  cc_conv = charset_to_ucs[cc_charset];
}

void
char_convert(byte **psb, uns *psl, byte **pdb, uns *pdl)
{
  byte *sb = *psb;
  byte *sbe = sb + *psl;
  byte *db = *pdb;
  byte *dbe = db + *pdl - 3;
  uns c;

  while (sb < sbe && db < dbe)
    {
      c = cc_conv[*sb++];
      if (!c)
	continue;
      PUT_UTF8(db, c);
    }

  *psb = sb;
  *psl = sbe - sb;
  *pdb = db;
  if (db < dbe)
    *pdl = dbe - db + 3;
  else
    *pdl = 0;
}
