/*
 *	Sherlock Gatherer -- Content-Encoding Decoders
 *
 *	(c) 1997 Martin Mares, <mj@atrey.karlin.mff.cuni.cz>
 */

#include <stdio.h>
#include <string.h>
#include <alloca.h>
#include <fcntl.h>
#include <unistd.h>

#include <zlib.h>

#include "gather.h"

#define DECOMP_BUFFER 16384

uns max_decode;				/* Maximal resulting file size */
static uns byte_count;			/* Byte counter */

static uns
dwrite(byte *buf, uns size, int fh)
{
  if (byte_count > max_decode)
    return 503;
  byte_count += size;
  if ((uns) write(fh, buf, size) != size)
    {
      log(L_ERROR "Decoder: Write error: %m");
      return 504;
    }
  return 0;
}

static int
decode_gzip(byte *name, int fh)
{
  gzFile z;
  byte *temp = alloca(DECOMP_BUFFER);
  int cnt;
  int err = 0;

  z = gzopen(name, "rb");
  if (!z)
    {
      log(L_ERROR "Unable to decompress %s: %m", name);
      return 501;
    }

  while ((cnt = gzread(z, temp, DECOMP_BUFFER)) > 0)
    if (err = dwrite(temp, cnt, fh))
      return err;

  if (cnt < 0)
    {
      byte *e = (byte *) gzerror(z, &err);
      if (err == Z_ERRNO)
	log(L_ERROR "%s: ungzip: %m", name);
      else
	log(L_ERROR "%s: ungzip: %s", name, e);
      err = 501;
    }

  gzclose(z);
  return err;
}

static int
decode_deflate(byte *name, int fh)
{
  z_stream zs;
  byte *inbuf = alloca(DECOMP_BUFFER);
  byte *outbuf = alloca(DECOMP_BUFFER);
  int fr, cnt;
  int err = 0;

  bzero(&zs, sizeof(zs));
  fr = open(name, O_RDONLY, 0);
  if (fr < 0)
    die("Data file %s missing: %m", name);
  if ((cnt = inflateInit(&zs)) != Z_OK)
    {
      log(L_ERROR "%s: Inflate error %d", cnt);
      err = 502;
      goto bye0;
    }

newout:
  zs.next_out = outbuf;
  zs.avail_out = DECOMP_BUFFER;
  for(;;)
    {
      if (!zs.avail_in)
	{
	  cnt = read(fr, inbuf, DECOMP_BUFFER);
	  if (cnt < 0)
	    die("Read error on %s: %m", name);
	  zs.next_in = inbuf;
	  zs.avail_in = cnt;
	}
      cnt = inflate(&zs, Z_NO_FLUSH);
      if (cnt == Z_STREAM_END)
	break;
      if (cnt != Z_OK)
	{
	  log(L_ERROR "%s: Inflate error %d", name, cnt);
	  err = 502;
	  goto bye;
	}
      if (zs.avail_out)
	{
	  if (err = dwrite(outbuf, zs.avail_out, fh))
	    goto bye;
	  goto newout;
	}
    }

bye:
  inflateEnd(&zs);
bye0:
  close(fr);
  return err;
}

int
c_e_decode(struct odes *o, int method)
{
  byte *fn = find_aval(o, '@');
  struct tempfile tf;
  int err;

  if (!fn)
    die("c_e_decode called on object with no data file");
  open_temp(&tf, TF_DECODE);
  byte_count = 0;

  switch (method)
    {
    case ET_GZIP:
      err = decode_gzip(fn, tf.fh);
      break;
    case ET_DEFLATE:
      err = decode_deflate(fn, tf.fh);
      break;
    default:
      return 500;
    }

  if (err == 503)
    {
      add_attr(o, NULL, '.', "Truncated to maxsize while decoding");
      err = 0;
    }

  prepend_attr(o, '@', tf.name);
  close(tf.fh);
  return err;
}
