/*
 *	Sherlock Gatherer Database Expiration Utility
 *
 *	(c) 1997 Martin Mares, <mj@atrey.karlin.mff.cuni.cz>
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/time.h>
#include <getopt.h>
#include <dirent.h>
#include <sys/stat.h>

#include "gather.h"

#include <lib/url.h>
#include <lib/fastbuf.h>

static byte *log_file;
static uns refresh_after = ~0, expire_after = ~0, expire_err_after = ~0, verbose, rem_empty;
static uns obj_count, queue_count, requeue_count, expire_count, filter_count;
static uns empty_dir_count;

uns show_forbidden;

static struct fastbuf *F;

static struct cfitem cfitems[] = {
  { "LogFile", CI_STRING, &log_file },
  { "RefreshAfter", CI_INT, &refresh_after },
  { "ExpireAfter", CI_INT, &expire_after },
  { "ExpireErrorsAfter", CI_INT, &expire_err_after },
  { "Verbose", CI_INT, &verbose },
  { NULL, CI_STOP, NULL }
};

struct loc_filter {
  struct loc_filter *next;
  regex *rx;
};

static struct loc_filter *first_lf, **last_lf = &first_lf;

static void
load_filter(byte *name)
{
  FILE *f;
  byte x[256];
  struct loc_filter *q;
  byte *c;

  if (! (f = fopen(name, "r")))
    die("Unable to open %s: %m", name);
  while (fgets(x, sizeof(x), f))
    {
      if (c = strchr(x, '\n'))
	*c = 0;
      q = xmalloc(sizeof(struct loc_filter));
      *last_lf = q;
      last_lf = &q->next;
      q->next = NULL;
      q->rx = rx_compile(x);
    }
  fclose(f);
}

static int
try_filter(byte *url)
{
  struct loc_filter *q;
  byte b1[MAX_URL_SIZE], b2[MAX_URL_SIZE], u0[MAX_URL_SIZE];
  struct url uu;
  int err;

  strcpy(u0, url);
  if (err = url_canon_split(u0, b1, b2, &uu))
    {
      log(L_ERROR "Error parsing `%s': %s", u0, err);
      return 0;
    }
  if (translate_url(&uu, u0, b1, b2) < 0)
    {
      if (verbose > 3)
	log(L_DEBUG "Translator said NO");
      return 0;
    }
  if (!okay_to_get(NULL, &uu))
    {
      if (verbose > 3)
	log(L_DEBUG "URL filters said NO");
      return 0;
    }

  for(q=first_lf; q; q=q->next)
    if (rx_match(q->rx, url))
      {
	if (verbose > 3)
	  log(L_DEBUG "User regex said NO");
	return 0;
      }

  return 1;
}

static void
kill_obj(struct urlrec *r, byte *url)
{
  byte name[64];
  struct odes *d;
  struct oattr *a;
  byte *x;

  if (r->oid < OID_MIN)
    goto delurl;
  mk_obj_name("obj", r->oid, name);
  d = obj_load(name);
  a = find_attr(d, 'U');
  while (a)
    {
      if (!strcmp(a->val, url))
	break;
      a = a->same;
    }
  if (!a)
    {
      log(L_ERROR "%s: object %08x doesn't contain this URL", url, r->oid);
    erex:
      obj_free(d);
      return;
    }
  if (!del_attr(d, a))
    {
      log(L_ERROR "%s: del_attr failed", url);
      goto erex;
    }
  if (find_aval(d, 'U'))
    {
      if (verbose)
	log(L_INFO "Object %08x has more references", r->oid);
      obj_write(name, d);
      obj_free(d);
      goto delurl;
    }
  if (x = find_aval(d, 'M'))
    {
      byte md5[MD5_SIZE];
      if (verbose > 5)
	log(L_DEBUG "Deleting MD5 %s", x);
      hex_to_md5(x, md5);
      if (!md5db_delete(md5))
	{
	  log(L_ERROR "Error while deleting MD5 %s for %s", x, url);
	  goto erex;
	}
    }
  obj_free(d);
  if (unlink(name) < 0)
    log(L_ERROR "unlink(%s) failed: %m", name);

delurl:
  if (!urldb_delete(url))
    log(L_ERROR "Error while deleting URL %s", url);
}

static void
requeue_obj(struct urlrec *r, byte *url)
{
  struct url u;
  int err;
  byte b1[MAX_URL_SIZE], b2[MAX_URL_SIZE];
  struct qhost *h;

  err = url_canon_split(url, b1, b2, &u);
  if (err)
    {
      log(L_ERROR "%s: %s", url, url_error(err));
      return;
    }
  h = find_host(u.protoid, u.host, u.port);
  if (!h)
    {
      log(L_ERROR "%s: Host %d:%s:%d not found", url, u.protoid, u.host, u.port);
      return;
    }
  enqueue_item(h, u.rest);
  enqueue_host(h);
  urldb_store(url, r->access|1, r->oid);
}

static void
expire_fill(void)
{
  byte buf[MAX_URL_SIZE];
  struct urlrec *r;
  uns now = time(NULL);
  uns age;

  if (!urldb_get_first(buf))
    return;
  do
    {
      if (!(r = urldb_lookup(buf)))
	die("%s: No data found", buf);
      obj_count++;
      if (r->access & 1)
	{
	  if (verbose > 5)
	    log(L_DEBUG "%s: queued", buf);
	  queue_count++;
	  continue;
	}
      if (r->access > now)
	{
	  log(L_ERROR "%s: Ante-dated (access=%d, now=%d)", buf, r->access, now);
	  continue;
	}
      age = now - r->access;
      if (!try_filter(buf))		/* Filtering out */
	{
	  if (verbose)
	    log(L_INFO "%s: Filtering out (age=%d)", buf, age);
	  filter_count++;
	  bputc(F, '-');
	  bputsn(F, buf);
	}
      else if (r->oid < OID_MIN)	/* It's an error */
	{
	  if (age >= expire_err_after)
	    {
	      if (verbose)
		log(L_INFO "%s [%d]: Expiring (age=%d)", buf, r->oid, age);
	      expire_count++;
	      bputc(F, '-');
	      bputsn(F, buf);
	    }
	  else if (verbose > 5)
	    log(L_DEBUG "%s [%d]: OK", buf, r->oid);
	}
      else if (age >= expire_after)
	{
	  if (verbose)
	    log(L_INFO "%s: Expiring (age=%d)", buf, age);
	  expire_count++;
	  bputc(F, '-');
	  bputsn(F, buf);
	}
      else if (age >= refresh_after)
	{
	  if (verbose)
	    log(L_INFO "%s: Requeued (age=%d)", buf, age);
	  requeue_count++;
	  bputc(F, '+');
	  bputsn(F, buf);
	}
      else if (verbose > 5)
	log(L_DEBUG "%s: OK", buf);
    }
  while (urldb_get_next(buf));
}

static void
expire_read(void)
{
  byte buf[MAX_URL_SIZE+4];
  struct urlrec *r;
  byte *url = buf+1;

  while (bgets(F, buf, sizeof(buf)))
    {
      if (verbose > 3)
	puts(buf);
      if (!(r = urldb_lookup(url)))
	die("%s: No data found", url);
      switch (buf[0])
	{
	case '+':
	  requeue_obj(r, url);
	  break;
	case '-':
	  kill_obj(r, url);
	  rem_empty = 1;
	  break;
	default:
	  die("%s: Unknown operation");
	}
    }
}

static uns
do_scan_dirs(byte *name, byte *p, byte *end)
{
  DIR *d;
  int empty = 1;
  struct dirent *e;
  struct stat st;

  *p = 0;
  d = opendir(name);
  if (!d)
    {
      log(L_ERROR "Unable to scan %s: %m", name);
      return 0;
    }
  while (e = readdir(d))
    {
      byte *n = e->d_name;
      int l = strlen(n);

      if (p + l + 2 >= end)
	{
	  log(L_ERROR "Path too long");
	  continue;
	}
      strcpy(p, n);
      if (stat(name, &st) < 0)
	{
	  log(L_ERROR "stat %s: %m", name);
	  goto notempty;
	}
      if (S_ISDIR(st.st_mode))
	{
	  if (!strcmp(n, ".") || !strcmp(n, ".."))
	    continue;
	  p[l] = '/';
	  if (do_scan_dirs(name, p+l+1, end))
	    {
	      p[l] = 0;
	      if (lstat(name, &st) < 0)
		{
		  log(L_ERROR "lstat %s: %m", name);
		  goto notempty;
		}
	      if (S_ISLNK(st.st_mode))
		goto notempty;
	      if (verbose)
		log(L_INFO "Removing dir %s", name);
	      if (rmdir(name))
		log(L_ERROR "Unable to remove %s: %m", name);
	      else
		empty_dir_count++;
	      continue;
	    }
	}
    notempty:
	empty = 0;
    }
  closedir(d);
  return empty;
}

static void
scan_empty_dirs(byte *base)
{
  byte name[256];

  strcpy(name, base);
  do_scan_dirs(name, name + strlen(base), name + sizeof(name) - 2);
}

static void
rem_empty_dirs(void)
{
  log(L_INFO "Scanning for empty directories");
  scan_empty_dirs("host/");
  scan_empty_dirs("obj/");
}

static void help(void) NONRET;

static void
help(void)
{
  fprintf(stderr, "Usage: expire [-v] [-x <extra-filter>] [-f <custom-list>] [-td]\n");
  exit(1);
}

static char options[] = "vx:f:td";

int
main(int argc, char **argv)
{
  int i;
  char *extra_filter = NULL;
  char *custom_list = NULL;
  int test_only = 0;

  initlog(argv[0]);
  cf_read("cf/expire", cfitems);
  open_log_file(log_file);
  db_init();
  url_filter_init();

  while ((i = getopt(argc, argv, options)) != -1)
    switch (i)
      {
      case 'v':
	verbose++;
	break;
      case 'x':
	extra_filter = optarg;
	break;
      case 'f':
	custom_list = optarg;
	break;
      case 't':
	test_only = 1;
	break;
      case 'd':
	rem_empty = 1;
	break;
      default:
	help();
      }
  if (optind != argc)
    help();

  log(L_INFO "Expire started");

  if (custom_list)
    {
      if (!(F = bopen(custom_list, O_RDONLY, 65536)))
	die("Cannot open input file: %m");
    }
  else
    {
      if (!(F = bopen("tmp/expire.tmp", O_CREAT | O_TRUNC | O_RDWR, 65536)))
	die("Cannot open tmp file: %m");
    }

  if (extra_filter)
    load_filter(extra_filter);

  gather_lock();
  queue_init();
  urldb_open();
  md5db_open();

  if (!custom_list)
    {
      expire_fill();
      bflush(F);
      bsetpos(F, 0);
      if (!test_only)
	{
	  log(L_INFO "Pass 2");
	  expire_read();
	  bclose(F);
	  unlink("tmp/expire.tmp");
	}
    }
  else if (!test_only)
    expire_read();

  if (rem_empty)
    rem_empty_dirs();

  queue_cleanup();
  urldb_close();
  md5db_close();
  gather_unlock();
  log(L_INFO "Expire done: seen %d objects, %d already queued, requeued %d,",
      obj_count, queue_count, requeue_count);
  log(L_INFO "filtered out %d, expired %d, removed %d empty dirs.",
      filter_count, expire_count, empty_dir_count);
  return 0;
}
