/*
 *	Sherlock Gatherer Control Utility
 *
 *	(c) 1997 Martin Mares, <mj@atrey.karlin.mff.cuni.cz>
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/time.h>

#include "gather.h"

#include <lib/url.h>

/* Verbosity level */

static int verbose;

/* Internalities of the queue manager */

extern list host_queue, idle_hosts;
extern int queue_fd;

/* Other magic variables */

uns show_forbidden;

/* Support functions */

static byte block[QUEUE_BLOCK_SIZE];

static void
get_block(uns at)
{
  if ((uns) lseek(queue_fd, at, SEEK_SET) != at)
    die("Queue lseek failed at %d: %m", at);
  if (read(queue_fd, block, QUEUE_BLOCK_SIZE) != QUEUE_BLOCK_SIZE)
    die("Queue read failed at %d: %m", at);
}

static struct url url;
static byte ubuf0[MAX_URL_SIZE], ubuf1[MAX_URL_SIZE], ubuf2[MAX_URL_SIZE], ubuf3[MAX_URL_SIZE];

static void
parse_url(byte *u)
{
  int err = url_canon_split(u, ubuf0, ubuf1, &url);
  if (err)
    die("Error parsing %s: %s", u, url_error(err));
}

/* List hosts and queued objects */

static void
show_host(struct qhost *h)
{
  printf("%s://%s:%d [%08x]\n", url_proto_names[h->protocol], h->name, h->port, h->host_id);
}

static void
show_host_queue(struct qhost *h)
{
  uns pos, next;

  show_host(h);
  putchar('\n');
  next = h->qf_pos;
  while (pos = next)
    {
      get_block(pos & ~(QUEUE_BLOCK_SIZE-1));
      pos &= (QUEUE_BLOCK_SIZE-1);
      next = (block[0] << 24) | (block[1] << 16) | (block[2] << 8) | block[3];
      while (block[pos])
	{
	  putchar('\t');
	  while (block[pos])
	    putchar(block[pos++]);
	  putchar('\n');
	  pos++;
	}
    }
}

static void
list_queue(byte *u)
{
  struct qhost *h;
 
  queue_init();
  if (u)
    {
      parse_url(u);
      h = find_host(url.protoid, url.host, url.port);
      if (!h)
	die("Host not found");
      show_host_queue(h);
    }
  else
    DO_FOR_ALL(h, host_queue)
    {
      show_host_queue(h);
      putchar('\n');
    }
}

static void
list_hosts(byte *u)
{
  struct qhost *h;
 
  queue_init();
  if (u)
    {
      parse_url(u);
      h = find_host(url.protoid, url.host, url.port);
      if (!h)
	die("Host not found");
      show_host(h);
    }
  else
    {
      DO_FOR_ALL(h, host_queue)
	show_host(h);
      DO_FOR_ALL(h, idle_hosts)
	{
	  fputs("[IDLE] ", stdout);
	  show_host(h);
	}
    }
}

static void
list_idles(void)
{
  struct qhost *h;

  queue_init();
  DO_FOR_ALL(h, idle_hosts)
    show_host(h);
}

static void
unqueue(byte *ur)
{
  struct qhost *h;
  struct qitem *a;
  struct urlrec *u;
  int err;

  parse_url(ur);
  gather_lock();
  urldb_open();
  queue_init();
  if (! (h = find_host(url.protoid, url.host, url.port)))
    log(L_ERROR "Unknown host");
  else
    {
      while (a = dequeue_item(h))
	{
	  printf("%s: ", a->text);
	  url.rest = a->text;
	  if (err = url_pack(&url, ubuf2))
	    die("url_pack: %s", url_error(err));
	  if (err = url_enescape(ubuf2, ubuf3))
	    die("url_enescape: %s", url_error(err));
	  if (u = urldb_lookup(ubuf3))
	    {
	      if (! (u->access & 1))
		printf("BAD DB RECORD\n");
	      else if (u->oid)
		{
		  printf("%08x\n", u->oid);
		  u->access &= ~1;
		  urldb_store(ubuf3, u->access, u->oid);
		}
	      else if (urldb_delete(ubuf3))
		printf("DELETED\n");
	      else
		printf("DELETE FAILED\n");
	    }
	  else
	    printf("NO DB ITEM\n");
	}
    }
  queue_cleanup();
  urldb_close();
  gather_unlock();
}

/* URLs */

static void
show_url0(byte *n, struct urlrec *z)
{
  time_t t = z->access & ~1;
  struct tm *tm;

  printf("%s%s ", n, (z->access & 1) ? " [QUEUED]" : "");
  if (!z->oid)
    ;
  else if (z->oid == 8888)
    printf("EMPTY ");
  else if (z->oid < OID_MIN)
    printf("Error %d ", z->oid);
  else
    printf("%08x ", z->oid);
  if (! (tm = localtime(&t)))
    die("Time parse error");
  printf ("(%02d-%02d-%02d %02d:%02d)\n",
	  tm->tm_mday, tm->tm_mon+1, tm->tm_year % 100,
	  tm->tm_hour, tm->tm_mon);
}

static void
show_url(byte *u)
{
  struct urlrec *z;
  byte *a;

  gather_lock();
  urldb_open();
  if (u)
    {
      z = urldb_lookup(u);
      if (!z)
	log(L_ERROR "%s: Not found", u);
      else
	show_url0(u, z);
    }
  else
    {
      a = urldb_get_first(ubuf0);
      while (a)
	{
	  if (! (z = urldb_lookup(a)))
	    die("URLDB inconsistent");
	  show_url0(a, z);
	  a = urldb_get_next(ubuf0);
	}
    }
  urldb_close();
  gather_unlock();
}

static void
del_url(byte *u)
{
  gather_lock();
  urldb_open();
  if (!urldb_delete(u))
    log(L_ERROR "%s: Not found", u);
  urldb_close();
  gather_unlock();
}

/* MD5 Database */

static void
show_md0(byte *M, struct md5rec *z)
{
  byte m[MD5_HEX_SIZE];

  md5_to_hex(M, m);
  printf("%s: %08x\n", m, z->oid);
}

static void
show_md5(byte *m)
{
  struct md5rec *z;
  byte *a;
  byte M[MD5_SIZE];

  if (m)
    hex_to_md5(m, M);
  gather_lock();
  md5db_open();
  if (m)
    {
      z = md5db_lookup(M);
      if (!z)
	log(L_ERROR "%s: Not found", m);
      else
	show_md0(M, z);
    }
  else
    {
      a = md5db_get_first();
      while (a)
	{
	  if (! (z = md5db_lookup(a)))
	    die("URLDB inconsistent");
	  show_md0(a, z);
	  a = md5db_get_next();
	}
    }
  md5db_close();
  gather_unlock();
}

static void
del_md5(byte *m)
{
  byte M[MD5_SIZE];

  hex_to_md5(m, M);
  gather_lock();
  md5db_open();
  if (!md5db_delete(M))
    log(L_ERROR "%s: Not found", m);
  md5db_close();
  gather_unlock();
}

/* URL's */

static void
filter_test(void)
{
  byte b[MAX_URL_SIZE];

  url_filter_init();

  if (verbose)
    {
      trace_subst = 1;
      trace_filter = 1;
    }

  while (fgets(b, sizeof(b), stdin))
    {
      byte *k = strchr(b, '\n');
      int rc;

      if (k)
	*k = 0;
      rc = url_canon_split(b, ubuf0, ubuf1, &url);
      if (rc)
	{
	  printf("- %s\n", url_error(rc));
	  continue;
	}
      rc = translate_url(&url, b, ubuf0, ubuf1);
      if (rc < 0)
	{
	  puts("- Subst failed");
	  continue;
	}
      if (rc)
	printf("> %s\n", b);
      if (okay_to_get(NULL, &url))
	puts("+ OK");
      else
	puts("- Forbidden");
    }
}

/* Main program */

static void
usage(void)
{
  fprintf(stderr, "Usage: gc [-v] <command>\n\
\n\
Available commands:\n\
\n\
-h [<url>]\tList known hosts [or host info]\n\
-f\t\tTest URL filtering\n\
-i\t\tList idle hosts\n\
-m\t\tList known MD5 sums\n\
-m <md5>\tList object matching given MD5 sum\n\
-M <md5>\tDelete from MD5 database\n\
-q [<url>]\tList queued objects [matching host, protocol and port of given URL]\n\
-Q <url>\tRemove all queued objects matching host, protocol and port of given URL\n\
-u\t\tList known URLs\n\
-u <url>\tList object matching given URL\n\
-U <url>\tDelete from URL database\n\
\n");
  exit(1);
}

int
main(int argc, char **argv)
{
  initlog(argv[0]);
  db_init();

again:
  if (argc < 2 || argv[1][0] != '-' || !argv[1][1] || argv[1][2])
    usage();
  
  switch (argv[1][1])
    {
    case 'h':
      if (argc == 2)
	list_hosts(NULL);
      else if (argc == 3)
	list_hosts(argv[2]);
      else
	usage();
      break;
    case 'i':
      if (argc != 2)
	usage();
      list_idles();
      break;
    case 'm':
      if (argc == 2)
	show_md5(NULL);
      else if (argc == 3)
	show_md5(argv[2]);
      else
	usage();
      break;
    case 'M':
      if (argc != 3)
	usage();
      del_md5(argv[2]);
      break;
    case 'q':
      if (argc == 2)
	list_queue(NULL);
      else if (argc == 3)
	list_queue(argv[2]);
      else
	usage();
      break;
    case 'Q':
      if (argc != 3)
	usage();
      unqueue(argv[2]);
      break;
    case 'u':
      if (argc == 2)
	show_url(NULL);
      else if (argc == 3)
	show_url(argv[2]);
      else
	usage();
      break;
    case 'U':
      if (argc != 3)
	usage();
      del_url(argv[2]);
      break;
    case 'f':
      if (argc != 2)
	usage();
      filter_test();
      break;
    case 'v':
      verbose++;
      argc--;
      argv++;
      goto again;
    default:
      usage();
    }
  return 0;
}
