/*
 *	Sherlock Gatherer -- Queue Management
 *
 *	(c) 1997 Martin Mares, <mj@atrey.karlin.mff.cuni.cz>
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <time.h>

#include "gather.h"

#include <lib/url.h>

/*** DEBUGGING ***/

#undef QUEUE_DEBUG

#ifdef QUEUE_DEBUG
#define DBG(x) log(L_DEBUG x)
#define DBG1(x,y) log(L_DEBUG x, y)
#define DBG2(x,y,z) log(L_DEBUG x, y, z)
#define DBG3(x,y,z,t) log(L_DEBUG x, y, z, t)
#else
#define DBG(x)
#define DBG1(x,y)
#define DBG2(x,y,z)
#define DBG3(x,y,z,t)
#endif

/*** FILE FUNCTIONS ***/

static uns
gw(FILE *f)
{
  return (fgetc(f) << 8) | fgetc(f);
}

static uns
gl(FILE *f)
{
  return (gw(f) << 16) | gw(f);
}

static void
pw(uns x, FILE *f)
{
  fputc(x >> 8, f);
  fputc(x, f);
}

static void
pl(uns x, FILE *f)
{
  pw(x >> 16, f);
  pw(x, f);
}

/*** HOST QUEUE ***/

list host_queue;			/* Main host queue (only those with some requests) */
static list active_hosts;		/* Hosts being processed (and dequeued) */
list idle_hosts;			/* Hosts with no queued requests */
static list waiting_hosts;		/* Hosts waiting to be serviced, sorted by wakeup time */
static struct qhost *host_hash[QUEUE_HASH_SIZE];
static struct qhost *free_list;		/* Fictive host holding list of free fragments */
static FILE *host_file;

static inline uns			/* Calculate hash value of given string */
hash(byte *name)
{
  uns h = strlen(name);

  while (*name)
    h = h*37 + *name++;
  return h & (QUEUE_HASH_SIZE - 1);
}

struct qhost *				/* Remove first host from the host queue */
dequeue_host(uns *waiting, u32 tim)
{
  struct qhost *h = HEAD(host_queue);
  struct qhost *w = HEAD(waiting_hosts);

  *waiting = 0;

  if (w->n.next && w->wait_until <= tim) /* Wait timed out => dequeue waiting host */
    {
      DBG1("Dequeued waiting host: <%s>", w->name);
      h = w;
    }
  else if (h->n.next)			/* Process normal queue */
    {
      DBG1("Dequeued host: <%s>", h->name);
    }
  else
    {
      if (w->n.next)			/* Just wait? */
	{
	  DBG1("Waiting until %d", w->wait_until);
	  *waiting = w->wait_until - tim;
	}
      return NULL;
    }

  rem_node(&h->n);
  add_tail(&active_hosts, &h->n);
  h->status = QHF_ACTIVE;
  return h;
}

struct qhost *				/* Find (possibly queued) host by its signature */
find_host(uns proto, byte *name, uns port)
{
  struct qhost *h;

  for(h = host_hash[hash(name)];
      h && (h->protocol != proto || h->port != port || strcmp(h->name, name));
      h = h->hash)
    ;
  return h;
}

static struct qhost *			/* Create new host entry */
new_host_entry(uns proto, byte *name, uns port, oid_t id)
{
  struct qhost *h = xmalloc(sizeof(struct qhost) + strlen(name));
  uns hh = hash(name);

  bzero(h, sizeof(struct qhost));
  h->hash = host_hash[hh];
  host_hash[hh] = h;
  h->host_id = id;
  h->port = port;
  h->protocol = proto;
  h->wait_until = 0;
  h->status = QHF_NEW;
  strcpy(h->name, name);
  return h;
}

struct qhost *				/* Create new host and make it active */
new_host(uns proto, byte *name, uns port, oid_t id)
{
  struct qhost *h = new_host_entry(proto, name, port, id);

  add_tail(&active_hosts, &h->n);	/* In active list, but marked QHF_NEW */
  return h;
}

static void				/* Write host entry at current position of host file */
write_host_entry(FILE *f, struct qhost *h)
{
  fputc(h->protocol, f);
  pw(h->port, f);
  pl(h->host_id, f);
  h->hf_pos = ftell(f);
  pl(h->qf_pos, f);
  pl(h->qf_last, f);
  fputs(h->name, f);
  fputc('\n', f);
}

static void				/* Rewrite host entry with new data */
flush_host_entry(struct qhost *h)
{
  DBG3("Flushing <%s> %x %x", h->name, h->qf_pos, h->qf_last);
  fseek(host_file, h->hf_pos, SEEK_SET);
  pl(h->qf_pos, host_file);
  pl(h->qf_last, host_file);
  if (h->status == QHF_ACTIVE_MODIFIED)
    h->status = QHF_ACTIVE;
}

static inline void			/* Schedule host entry for re-writing */
touch_host_entry(struct qhost *h)
{
  DBG1("Touching <%s>", h->name);
  if (h->status == QHF_ACTIVE || h->status == QHF_ACTIVE_MODIFIED)
    h->status = QHF_ACTIVE_MODIFIED;
  else if (h->status != QHF_NEW)
    flush_host_entry(h);
}

static inline void			/* Write free list if it's been modified */
flush_free_list(void)
{
  if (free_list->status == QHF_ACTIVE_MODIFIED)
    flush_host_entry(free_list);
}

static void				/* Flush all modified host entries */
flush_host_cache(void)
{
  struct qhost *h;

  DO_FOR_ALL(h, active_hosts)
    if (h->status == QHF_ACTIVE_MODIFIED)
      flush_host_entry(h);
  flush_free_list();
}

void					/* Add active host at the end of the host queue */
enqueue_host(struct qhost *h)
{
  DBG2("enqueue_host(%s) (status %d)", h->name, h->status);
  flush_free_list();
  switch (h->status)
    {
    case QHF_NEW:			/* New record -> append it */
      fseek(host_file, 0, SEEK_END);
      write_host_entry(host_file, h);
      rem_node(&h->n);			/* Was in the active queue... */
      break;
    case QHF_ACTIVE_MODIFIED:		/* Modified -> flush it */
      flush_host_entry(h);
      /* Fall-thru */
    case QHF_ACTIVE:
    case QHF_EMPTY:
      rem_node(&h->n);
      break;
    case QHF_QUEUED:
    case QHF_WAITING:
      return;
    case QHF_LOADED:
      break;
    default:
      die("enqueue_host: invalid status=%d", h->status);
    }

  if (h->qf_pos)
    {
      h->status = QHF_QUEUED;
      add_tail(&host_queue, &h->n);
    }
  else					/* No queued items => move to idle queue */
    {
      h->status = QHF_EMPTY;
      add_tail(&idle_hosts, &h->n);
      DBG1("%s is idle...", h->name);
    }
  h->wait_until = 0;
}

void					/* Insert active record to the wait queue */
wait_host(struct qhost *h, uns until)
{
  struct qhost *r, *n;

  DBG2("%s waits until %d", h->name, until);
  if (h->status == QHF_ACTIVE_MODIFIED)
    flush_host_entry(h);
  if (h->status != QHF_ACTIVE)
    die("wait_host: bad status");
  rem_node(&h->n);
  h->status = QHF_WAITING;
  h->wait_until = until;
  for(r = (void *) &waiting_hosts.head;
      (n = (void *) r->n.next) && n->n.next && n->wait_until < until;
      r = n);
    ;
  insert_node(&h->n, &r->n);
}

/* ITEM QUEUE */

int queue_fd;				/* File handle for the queue file */

struct qbuf {
  node n;
  uns pos;
  uns dirty;				/* Data have been modified */
  uns next;				/* File position of next block */
  byte data[QUEUE_BLOCK_SIZE];
};

static list queue_cache;
static uns cache_slots;			/* Current number of used cache slots */
static uns queue_size;

static void				/* Flush possibly modified buffer to disk */
qbuf_flush(struct qbuf *b)
{
  DBG1("Flushing %x", b->pos);
  if (b->dirty)
    {
      DBG(":: dirty");
      b->data[0] = b->next >> 24;
      b->data[1] = b->next >> 16;
      b->data[2] = b->next >> 8;
      b->data[3] = b->next;
      if (lseek(queue_fd, b->pos, SEEK_SET) < 0)
	die("Queue seek failed: %m");
      if (write(queue_fd, b->data, QUEUE_BLOCK_SIZE) != QUEUE_BLOCK_SIZE)
	die("Queue write failed: %m");
      b->dirty = 0;
    }
}

static struct qbuf *			/* Create new cache slot, possibly flushing another one */
qbuf_new(uns pos)
{
  struct qbuf *b;

  if (cache_slots < QUEUE_CACHE_SIZE)
    {
      b = xmalloc(sizeof(struct qbuf));
      cache_slots++;
    }
  else
    {
      b = HEAD(queue_cache);
      rem_node(&b->n);
      qbuf_flush(b);
    }
  add_tail(&queue_cache, &b->n);
  b->pos = pos;
  b->dirty = 0;
  return b;
}

static struct qbuf *			/* Get queue file block containing specified position */
qbuf_get(uns pos)
{
  struct qbuf *b;
  uns pos0 = pos & ~(QUEUE_BLOCK_SIZE - 1);

  if (pos >= queue_size)
    die("qbuf_get: out of range (%d vs. %d)", pos, queue_size);

  DO_FOR_ALL(b, queue_cache)
    if (b->pos == pos0)
      {
	DBG1(":: %x found in cache", pos0);
	rem_node(&b->n);
	add_tail(&queue_cache, &b->n);
	return b;
      }

  b = qbuf_new(pos0);
  DBG1(":: reading block %x", pos0);
  if (lseek(queue_fd, pos0, SEEK_SET) < 0)
    die("Queue seek: %m");
  if (read(queue_fd, b->data, QUEUE_BLOCK_SIZE) != QUEUE_BLOCK_SIZE)
    die("Queue read: %m");
  b->next = (b->data[0] << 24) | (b->data[1] << 16) | (b->data[2] << 8) | b->data[3];
  return b;
}

static struct qbuf *			/* Allocate new block in the queue */
qblock_alloc(void)
{
  struct qbuf *b;
  uns pos;

  if (pos = free_list->qf_pos)
    {
      b = qbuf_get(pos);
      free_list->qf_pos = b->next;
      touch_host_entry(free_list);
      DBG1("Got block %x from free list", b->pos);
    }
  else
    {
      pos = queue_size;
      queue_size += QUEUE_BLOCK_SIZE;
      b = qbuf_new(pos);
      bzero(b->data, QUEUE_BLOCK_SIZE);
      DBG1("Got block %x from end of file", b->pos);
    }
  b->next = 0;
  b->dirty = 1;
  return b;
}

static void				/* Return block to the block pool */
qblock_free(struct qbuf *b)
{
  DBG1("Returning block %x to the pool", b->pos);
  bzero(b->data, QUEUE_BLOCK_SIZE);
  b->next = free_list->qf_pos;
  free_list->qf_pos = b->pos;
  touch_host_entry(free_list);
  b->dirty = 1;
}

static void				/* Flush all buffered blocks */
qcache_flush(void)
{
  struct qbuf *b;

  DO_FOR_ALL(b, queue_cache)
    qbuf_flush(b);
}

static struct qitem *qi = NULL;

struct qitem *				/* Get one item without dequeueing it first */
peek_item(struct qhost *h)
{
  uns pos;
  struct qbuf *b;
  byte *d;

  if (!qi)
    qi = xmalloc(sizeof(struct qitem) + MAX_URL_SIZE - 1);

  DBG1("peek_item(%s)", h->name);
  while (pos = h->qf_pos)
    {
      DBG1("...peeking at %x", pos);
      b = qbuf_get(pos);
      pos &= (QUEUE_BLOCK_SIZE - 1);
      if (b->data[pos])
	{
	  d = qi->text;
	  while (*d = b->data[pos])
	    d++, pos++;
	  qi->aux = b->pos + pos + 1;
	  return qi;
	}
      pos = b->next;
      h->qf_pos = pos;
      qblock_free(b);
      if (!pos)
	h->qf_last = 0;
      touch_host_entry(h);
    }
  return NULL;
}

struct qitem *				/* Fetch item and dequeue it */
dequeue_item(struct qhost *h)
{
  struct qitem *i = peek_item(h);

  if (i)
    {
      h->qf_pos = i->aux;
      touch_host_entry(h);
    }
  return i;
}

void					/* Add new item to the queue */
enqueue_item(struct qhost *h, byte *name)
{
  struct qbuf *b;
  uns last;
  uns len = strlen(name) + 1;

  DBG2("enqueue_item(%s,%s)", h->name, name);
  if (!h->qf_last)
    {
      b = qblock_alloc();
      last = 4;
      DBG1("...allocated first block at %x", b->pos);
    }
  else
    {
      b = qbuf_get(h->qf_last);
      last = h->qf_last & (QUEUE_BLOCK_SIZE - 1);
      DBG2("...block at %x, last=%x", b->pos, last);
      if (last + len >= QUEUE_BLOCK_SIZE)
	{
	  struct qbuf *o = b;
	  b = qblock_alloc();		/* We rely on this not flushing the old qbuf [see Appendix D] */
	  o->next = b->pos + 4;
	  o->dirty = 1;
	  last = 4;
	  DBG2("...overflow, new block at %x, last=%x", b->pos, last);
	}
    }
  strcpy(b->data + last, name);
  b->data[last+len-1] = 0;
  if (!h->qf_pos)
    h->qf_pos = b->pos + last;
  h->qf_last = b->pos + last + len;
  b->dirty = 1;
  touch_host_entry(h);
}

/* INITIALIZATION AND CLEANUP */

static void				/* Load list of all hosts from the host file */
load_host_list(void)
{
  FILE *f;

  DBG("Loading host list");
  if (! (host_file = f = fopen(HOST_FILE_NAME, "r+")))
    {
      log(L_INFO "Host database doesn't exist, creating new one");
      host_file = f = fopen(HOST_FILE_NAME, "w");
      if (!f)
	die("Host database creation failed: %m");
      free_list = new_host_entry(0, "", 0, 0);
      write_host_entry(f, free_list);
    }
  else
    {
      for(;;)
	{
	  int proto;
	  uns port, hpos, qpos, lpos;
	  oid_t id;
	  struct qhost *h;
	  byte name[MAX_URL_SIZE];
	  byte *k;
		  
	  proto = fgetc(f);
	  if (proto == EOF)
	    break;
	  port = gw(f);
	  id = gl(f);
	  hpos = ftell(f);
	  qpos = gl(f);
	  lpos = gl(f);
	  if (!fgets(name, MAX_URL_SIZE, f) || ! (k = strchr(name, '\n')))
	    die("Malformed host file entry at %d", ftell(f));
	  *k = 0;
	  h = new_host_entry(proto, name, port, id);
	  h->qf_pos = qpos;
	  h->qf_last = lpos;
	  h->hf_pos = hpos;
	  h->status = QHF_LOADED;
	  if (!h->port && !h->protocol)
	    free_list = h;
	  else
	    enqueue_host(h);
	}
    }
  if (!free_list)
    die("Malformed host file: missing free list entry");
}

static void
flush_host_list(FILE *f, list *l)
{
  struct qhost *h;

  DO_FOR_ALL(h, *l)
    write_host_entry(f, h);
}

static void				/* Dump all known hosts to the host file */
flush_host_db(void)
{
  FILE *f;

  DBG("Flushing host database");
  fclose(host_file);
  unlink(HOST_BACK_NAME);
  if (rename(HOST_FILE_NAME, HOST_BACK_NAME) < 0)
    die("Cannot rename " HOST_FILE_NAME " to " HOST_BACK_NAME ": %m");
  f = fopen(HOST_FILE_NAME, "w");
  if (!f)
    die("Cannot write host file: %m");
  write_host_entry(f, free_list);
  flush_host_list(f, &active_hosts);
  flush_host_list(f, &host_queue);
  flush_host_list(f, &waiting_hosts);
  flush_host_list(f, &idle_hosts);
  fclose(f);
  unlink(HOST_BACK_NAME);
}

void					/* Initialize the queueing mechanism */
queue_init(void)
{
  DBG("queue_init");
  init_list(&host_queue);
  init_list(&active_hosts);
  init_list(&idle_hosts);
  init_list(&waiting_hosts);
  init_list(&queue_cache);
  load_host_list();
  free_list->status = QHF_ACTIVE;
  if ((queue_fd = open(QUEUE_FILE_NAME, O_RDWR | O_CREAT, 0666)) < 0)
    die("Unable to open queue file: %m");
  queue_size = lseek(queue_fd, 0, SEEK_END);
}

void					/* Cleanup of the queue machinery */
queue_cleanup(void)
{
  DBG("queue_cleanup");
  qcache_flush();
  flush_host_db();
  close(queue_fd);
}

void					/* Write out all modified records */
queue_sync(void)
{
  DBG("queue_sync");
  qcache_flush();
  flush_host_cache();
  fflush(host_file);
}
