/*
 *	Sherlock Gatherer -- Filtering
 *
 *	(c) 1997, 1998 Martin Mares, <mj@atrey.karlin.mff.cuni.cz>
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "gather.h"

#include <lib/string.h>
#include <lib/url.h>

struct ufilt {
  struct ufilt *next;
  uns proto;				/* 0=any */
  byte *host;
  uns port;				/* 0=any */
  byte *path;
  int allow;
};

static struct ufilt *uf_first = NULL;
static struct ufilt **uf_last = &uf_first;

struct uregex {
  struct uregex *next;
  byte *from, *to;
  regex *rx;
};

static struct uregex *ur_first = NULL;
static struct uregex **ur_last = &ur_first;

int trace_filter, trace_robots, allow_relative, trace_subst;
byte *robot_name = "sherlock";

static byte *
url_filter_add(struct cfitem *c, byte *l)
{
  byte *fields[4];
  struct ufilt *u = xmalloc(sizeof(struct ufilt));
  char *end;
  byte *any = "*";

  if (wordsplit(l, fields, 4) != 4)
    return "Invalid URL filter";
  if (!strcmp(fields[0], any))
    u->proto = 0;
  else if (! (u->proto = identify_protocol(fields[0])))
    return "Unknown protocol";
  u->host = stralloc(fields[1]);
  if (!strcmp(fields[2], any))
    u->port = 0;
  else
    {
      u->port = strtol(fields[2], &end, 0);
      if (end && *end || !u->port)
	return "Invalid port number";
    }
  u->path = stralloc(fields[3]);
  u->allow = c->name[0] == 'A';
  *uf_last = u;
  uf_last = &u->next;
  u->next = NULL;
  return NULL;
}

static byte *
url_regex_add(struct cfitem *c, byte *l)
{
  byte *fields[2];
  struct uregex *u = xmalloc(sizeof(struct uregex));

  if (wordsplit(l, fields, 2) != 2)
    return "Invalid URL substitution";
  u->from = stralloc(fields[0]);
  u->to = stralloc(fields[1]);
  u->rx = rx_compile(u->from);
  *ur_last = u;
  ur_last = &u->next;
  u->next = NULL;
  return NULL;
}

static struct cfitem cfitems[] = {
  { "Trace", CI_INT, &trace_filter },
  { "ShowRobots", CI_INT, &trace_robots },
  { "Allow", CI_FUNCTION, url_filter_add },
  { "Deny", CI_FUNCTION, url_filter_add },
  { "AllowRelative", CI_INT, &allow_relative },
  { "Subst", CI_FUNCTION, url_regex_add },
  { "TraceSubst", CI_INT, &trace_subst },
  { NULL, CI_STOP, NULL }
};

void
url_filter_init(void)
{
  cf_read("cf/url_filter", cfitems);
}

static inline int
match_single(struct ufilt *f, struct url *u)
{
  return (!f->proto || f->proto == u->protoid)
    && (!f->port || f->port == u->port)
    && match_pattern_nocase(f->host, u->host)
    && match_pattern(f->path, u->rest);
}

static uns
filter_url(struct url *u)
{
  struct ufilt *f;

  for(f=uf_first; f; f=f->next)
    {
      if (match_single(f, u))
	return f->allow;
    }
  return 1;
}

static uns
filter_dynamic(struct odes *d, struct url *u, uns type)
{
  struct oattr *a = find_attr(d, type);

  while (a)
    {
      if (!strncmp(u->rest, a->val, strlen(a->val)))
	return 0;
      a = a->same;
    }
  return 1;
}

static int
filter_rel(struct url *u)
{
  if (allow_relative)
    return 1;
  switch (u->protoid)
    {
    case URL_PROTO_HTTP:
    case URL_PROTO_FTP:
      break;
    default:
      return 1;
    }
  if (strchr(u->host, '.'))
    return 1;
  return 0;
}

uns
okay_to_get(struct odes *d, struct url *u)
{
  byte *msg = NULL;

  if (trace_filter)
    {
      byte url[MAX_URL_SIZE];
      int err = url_pack(u, url);
      if (err)
	log(L_ERROR "URL filter: URL error %d", err);
      log(L_DEBUG "URL filter: examining %s", url);
    }

  if (!filter_rel(u))
    msg = "relative";
  else if (!filter_url(u))
    msg = "global";
  else if (d)				/* Host description available? */
    {
      if (!filter_dynamic(d, u, 'F'))
	msg = "static local";
      else if (!filter_dynamic(d, u, 'f'))
	msg = "dynamic local";
    }

  if (msg)
    {
      if (trace_filter)
	log(L_DEBUG "Rejected by %s filter", msg);
      return 0;
    }
  else
    {
      if (trace_filter)
	log(L_DEBUG "Accepted");
      return 1;
    }
}

void					/* Convert downloaded robot file to access lists */
parse_robot_file(struct odes *h, struct odes *r)
{
  FILE *f;
  byte *name, *url;
  byte buf[256];
  byte *i, *j, *k;
  int itsme = 0;
  struct oattr *oa = NULL;

  set_attr(h, 'f', NULL);		/* Remove all old items */

  if (r && find_aval(r, 'X'))		/* Robot file is a redirect */
    {
      log(L_ERROR "%s: Robot file is a redirect (-> %s)", find_aval(r, 'U'), find_aval(r, 'R'));
      r = NULL;
    }

  if (!r)				/* No robot file found */
    {
      if (find_aval(h, 'R'))
	log(L_INFO "Access list for %s has ceased to exist", find_aval(h, 'H'));
      return;
    }

  url = find_aval(r, 'U');
  if (! (name = find_aval(r, '@')) || ! (f = fopen(name, "r")))
    die("parse_robot_file: No contents found");
  if (trace_robots)
    log(L_DEBUG "Parsing robot file %s", url);
  while (fgets(buf, sizeof(buf), f))
    {
      if (i = strchr(buf, '\n'))
	{
	  *i = 0;
	  if (i != buf && i[-1] == '\r')
	    i[-1] = 0;
	}
      if (buf[0] == '#' || !buf[0])
	continue;
      i = buf;
      while (Cspace(*i))
	i++;
      if (!*i)				/* Record separator */
	{
	  itsme = 0;
	  continue;
	}
      for(j=i; *j && !Cspace(*j) && *j !=':'; j++)	/* Find the argument */
	;
      if (*j)
	{
	  if (*j == ':')
	    *j++ = 0;
	  while (Cspace(*j))
	    *j++ = 0;
	}
      for(k=j; *k && !Cspace(*k); k++)	/* Chop off trailing blanks */
	;
      *k = 0;
      if (!strcasecmp(buf, "User-Agent"))
	{
	  if (!strncasecmp(robot_name, j, strlen(j)) || !strcmp(j, "*"))
	    itsme = 1;
	}
      else if (!strcasecmp(buf, "Disallow"))
	{
	  if (itsme)
	    {
	      if (!*j)			/* Allow all */
		{
		  if (trace_robots)
		    log(L_DEBUG "Resetting...");
		  set_attr(h, 'f', NULL);
		  oa = NULL;
		}
	      else			/* Disallow something */
		{
		  if (trace_robots)
		    log(L_DEBUG "Disallowing %s", j);
		  oa = add_attr(h, oa, 'f', j);
		}
	    }
	}
      else if ((i = strchr(buf, '<')) && strchr(i+1, '>'))
	{
	  log(L_ERROR "%s: Seems to be a HTML page", url);
	  break;
	}
      else
	log(L_ERROR "%s: Unknown keyword %s", url, buf);
    }
  fclose(f);
}

int
translate_url(struct url *u, byte *url, byte *buf1, byte *buf2)
{
  struct uregex *r;
  byte dest[MAX_URL_SIZE], src[MAX_URL_SIZE];
  byte *ss = url;
  byte *dd = dest;
  int match = 0;

  for(r=ur_first; r; r=r->next)
    {
      int a;
      a = rx_subst(r->rx, r->to, ss, dd, MAX_URL_SIZE);
      if (a < 0)
	{
	  log(L_ERROR "%s: Regex replace failed on %s", r->from, ss);
	  return -1;
	}
      if (a)
	{
	  byte *x = (match++) ? ss : src;
	  if (trace_subst)
	    log(L_DEBUG "%s -> %s", ss, dd);
	  ss = dd;
	  dd = x;
	}
    }
  if (match)
    {
      int err;
      if (!strcmp(ss, "DROP"))
	{
	  if (show_forbidden)
	    log(L_INFO "%s dropped by subst filter", url);
	  return -1;
	}
      if (   (err = url_canon_split(ss, buf1, buf2, u))
	  || (err = url_pack(u, dd))
	  || (err = url_enescape(dd, url)))
	{
	  log(L_ERROR "Subst %s -> %s failed: %s", url, ss, url_error(err));
	  return -1;
	}
      if (trace_subst)
	log(L_DEBUG "Replaced by %s", url);
    }
  return match;
}
