/*
 *	Sherlock HTTP Downloader
 *
 *	(c) 1997 Martin Mares, <mj@atrey.karlin.mff.cuni.cz>
 */

/*
 *	Calling convention:
 *
 *	Command line: <URL> [<if-mod-since> [<use-old-HTTP>]]
 *
 *	fd[1] = object header
 *	fd[2] = log file
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/signal.h>
#include <netdb.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <unistd.h>
#include <stdarg.h>
#include <sys/time.h>

#include <lib/lib.h>
#include <lib/string.h>
#include <lib/url.h>

/* Configuration parameters */

static byte *filter_add(struct cfitem *, byte *);

static uns max_hdr_lines = 1000;
static uns max_body_bytes = 1048576;
static byte *local_admin = NULL;
static byte *referer = NULL;
static byte *acc_charset = NULL;
static byte *acc_encoding = NULL;
static byte *acc_lang = NULL;
static byte *acc_types = NULL;
static uns if_modified_since = 0;
static uns use_old_http = 0;
static uns connect_timeout = 60;
static uns header_timeout = 60;
static uns body_timeout = 600;
static byte *proxy = NULL;
static uns proxy_port = 80;
static uns trace = 0;
static uns length_checks = 1;
static int dump_to_fd = -1;

static struct cfitem cfitems[] = {
  { "MaxHeaderLines", CI_INT, &max_hdr_lines },
  { "MaxBodyLength", CI_INT, &max_body_bytes },
  { "From", CI_STRING, &local_admin },
  { "Referer", CI_STRING, &referer },
  { "AcceptCharset", CI_STRING, &acc_charset },
  { "AcceptEncoding", CI_STRING, &acc_encoding },
  { "AcceptLanguage", CI_STRING, &acc_lang },
  { "AcceptTypes", CI_STRING, &acc_types },
  { "ConnectTimeout", CI_INT, &connect_timeout },
  { "HeaderTimeout", CI_INT, &header_timeout },
  { "BodyTimeout", CI_INT, &body_timeout },
  { "UseProxy", CI_STRING, &proxy },
  { "ProxyPort", CI_INT, &proxy_port },
  { "Trace", CI_INT, &trace },
  { "Filter", CI_FUNCTION, filter_add },
  { "LengthChecks", CI_INT, &length_checks },
  { "DumpToFD", CI_INT, &dump_to_fd },
  { NULL, CI_STOP, NULL }
};

/* Other globals */

static uns expected_length = 0;
static uns is_chunked = 0;

/* Error result functions */

static void invalid(byte *, ...) NONRET;
static void retry(byte *, ...) NONRET;

static void
invalid(byte *msg, ...)
{
  va_list args;

  va_start(args, msg);
  putchar('-');
  vprintf(msg, args);
  putchar('\n');
  if (trace)
    {
      byte buf[256];
      vsprintf(buf, msg, args);
      log(L_DEBUG "Fatal error: %s", buf);
    }
  exit(1);
}

static void
retry(byte *msg, ...)
{
  va_list args;

  va_start(args, msg);
  putchar('?');
  vprintf(msg, args);
  putchar('\n');
  if (trace)
    {
      byte buf[256];
      vsprintf(buf, msg, args);
      log(L_DEBUG "Retrying: %s", buf);
    }
  exit(1);
}

static void
timeout(int unused)
{
  retry("0102 HTTP timeout");
}

/* Set up a network connection */

static int
http_connect(byte *host, uns port)
{
  byte *name;
  struct hostent *h;
  struct sockaddr_in rem;
  int sock;

  name = host;
  if (proxy)
    {
      name = proxy;
      port = proxy_port;
    }
  h = gethostbyname(name);
  if (!h)
    {
      switch (h_errno)
	{
	case HOST_NOT_FOUND:
	case NO_ADDRESS:
	  invalid("0103 Host doesn't exist");
	case TRY_AGAIN:
	  retry("0104 DNS timeout");
	default:
	  invalid("0105 Unrecoverable DNS error");
	}
    }
  rem.sin_family = AF_INET;
  rem.sin_port = htons(port);
  memcpy(&rem.sin_addr.s_addr, h->h_addr, sizeof(rem.sin_addr.s_addr));
  if (trace)
    log(L_DEBUG "Opening connection to %s (%08x), port %d", name, ntohl(rem.sin_addr.s_addr), port);

  sock = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
  if (!sock)
    die("No socket (%m)");
  alarm(connect_timeout);
  if (connect(sock, (struct sockaddr *) &rem, sizeof(struct sockaddr_in)) < 0)
    retry("0107 Connect failed (%m)");

  return sock;
}

/* Sending of single header */

static void
sendhdr(FILE *f, byte *txt, ...)
{
  byte buf[1024];
  va_list args;

  va_start(args, txt);
  vsprintf(buf, txt, args);
  if (trace)
    log(L_DEBUG "> %s", buf);
  fputs(buf, f);
  fputs("\r\n", f);
}

/* Date formatting and parsing functions */

static void
http_form_date(byte *buf, time_t time)
{
  struct tm *tm;

  tm = gmtime(&time);
  strftime(buf, 64, "%a, %d %b %Y %H:%M:%S GMT", tm);
}

static int
validate(byte *a, byte *p)
{
  while (*p && *a)
    {
      switch (*p)
	{
	case '.':
	  break;
	case '$':
	  if (*a == ' ')
	    break;
	  /* Fall-thru */
	case '#':
	  if (!Cdigit(*a))
	    return 0;
	  break;
	case 'a':
	  if (!Clower(*a))
	    return 0;
	  break;
	case 'A':
	  if (!Cupper(*a))
	    return 0;
	  break;
	case '@':
	  if (!Calpha(*a))
	    return 0;
	  break;
	case '*':
	  return 1;
	default:
	  if (*p != *a)
	    return 0;
	}
      p++;
      a++;
    }
  return (*p == *a);
}

static int
twodig(byte *p)
{
  if (p[0] == ' ')
    return p[1] - '0';
  else
    return (p[0] - '0')*10 + p[1] - '0';
}

static int
fourdig(byte *p)
{
  return twodig(p)*100 + twodig(p+2);
}

static byte short_months[] = "JanFebMarAprMayJunJulAugSepOctNovDec";

static int
find_month(byte *p)
{
  byte *z = short_months;
  int m = 0;

  while (*z)
    {
      m++;
      if (p[0] == z[0] && p[1] == z[1] && p[2] == z[2])
	return m;
      z += 3;
    }
  return 0;
}

static uns
http_parse_date(byte *buf)
{
  byte *p = buf;
  byte *q;
  struct tm tm;
  int m, y;

  bzero(&tm, sizeof(tm));
#ifdef STRICT_RFC822_CHECK
  if (validate(p, "Aaa, ## Aaa #### ##:##:## GMT"))
#else
  if (validate(p, "Aaa, ## Aaa #### ##:##:##*"))
#endif
    {					/* RFC 822/1123 */
      tm.tm_mday = twodig(p+5);
      m = find_month(p+8);
      if (!m)
	goto unk;
      tm.tm_mon = m - 1;
      y = fourdig(p+12);
      if (y < 1980)
	goto unk;
      tm.tm_year = y - 1900;
      tm.tm_hour = twodig(p+17);
      tm.tm_min = twodig(p+20);
      tm.tm_sec = twodig(p+23);
      goto ook;
    }
  if (q = strchr(p, ','))
    {
      int flag = 0;
      if (validate(q, ", ##-Aaa-## ##:##:## GMT"))	/* RFC 850 */
        flag = 1;
      else if (validate(q, ", #-Aaa-## ##:##:## GMT"))	/* Incorrectly implemented RFC 850 */
	flag = 2;
      if (flag)
	{
	  tm.tm_mday = twodig(q+2);
	  if (flag == 2)
	    q--;
	  m = find_month(q+5);
	  if (!m)
	    goto unk;
	  tm.tm_mon = m - 1;
	  tm.tm_year = twodig(q+9);
	  if (tm.tm_year < 76)
	    tm.tm_year += 100;
	  tm.tm_hour = twodig(q+12);
	  tm.tm_min = twodig(q+15);
	  tm.tm_sec = twodig(q+18);
	  goto ook;
	}
    }
  if (validate(p, "Aaa Aaa $# ##:##:## ####"))
    {					/* ANSI C asctime() */
      m = find_month(p+4);
      if (!m)
	goto unk;
      tm.tm_mon = m - 1;
      tm.tm_mday = twodig(p+8);
      tm.tm_hour = twodig(p+11);
      tm.tm_min = twodig(p+14);
      tm.tm_sec = twodig(p+17);
      y = fourdig(p+20);
      if (y < 1980)
	goto unk;
      tm.tm_year = y - 1900;
      goto ook;
    }

unk:
  log(L_WARN "Unable to parse date `%s'", buf);
  return 0;

ook:
  m = timegm(&tm);
  if (m == (time_t) -1)
    goto unk;
  return m;
}

/* Sending of all required headers */

static void
send_header(FILE *f, byte *host, uns port, byte *rest)
{
  byte buf[1536];
  byte *k;

  strcpy(buf, "GET ");
  k = buf + 4;
  if (proxy)
    {
      k += sprintf(k, "http://%s", host);
      if (port != 80)
	k += sprintf(k, ":%d", port);
    }
  k += sprintf(k, "%s", rest);
  if (!use_old_http)
    k += sprintf(k, " HTTP/1.1");
  if (trace)
    {
      *k = 0;
      log(L_DEBUG "Sending %s", buf);
    }
  *k++ = '\r';
  *k++ = '\n';
  *k = 0;
  fputs(buf, f);

  if (!use_old_http)
    {
      if (port == 80)
	sendhdr(f, "Host: %s", host);
      else
	sendhdr(f, "Host: %s:%d", host, port);
      if (acc_types)
	sendhdr(f, "Accept: %s", acc_types);
      if (acc_charset)
	sendhdr(f, "Accept-Charset: %s", acc_charset);
      if (acc_encoding)
	sendhdr(f, "Accept-Encoding: %s", acc_encoding);
      if (acc_lang)
	sendhdr(f, "Accept-Language: %s", acc_lang);
      sendhdr(f, "Connection: close");
      if (local_admin)
	sendhdr(f, "From: %s", local_admin);
      if (referer)
	sendhdr(f, "Referer: %s", referer);
      if (if_modified_since)
	{
	  byte buf[32];
	  http_form_date(buf, if_modified_since);
	  sendhdr(f, "If-Modified-Since: %s", buf);
	}
      sendhdr(f, "User-Agent: sherlock/" SHER_VER " httpget/" SHER_VER);
      fputs("\r\n" ,f);
    }
  fflush(f);
}

/* Receiving of all headers */

struct hdr {
  struct hdr *next;
  byte *contents;
  byte name[1];
};

static struct hdr *firsthdr = NULL;
static struct hdr *resphdr = NULL;
static uns additional_hdr = 0;

static void
hdrflush(byte *x)
{
  byte *f;
  struct hdr *h;

  if (!*x)
    return;
  if (trace)
    log(L_DEBUG "%s %s", additional_hdr ? "<<" : "<", x);
  h = xmalloc(sizeof(struct hdr) + strlen(x));
  strcpy(h->name, x);
  if (!resphdr)
    {
      resphdr = h;
      f = NULL;
    }
  else
    {
      h->next = firsthdr;
      firsthdr = h;
      f = h->name;
      while (*f && *f != ' ' && *f != '\t')
	f++;
      while (*f == ' ' || *f == '\t')
	*f++ = 0;
    }
  h->contents = f;
}

static void
recv_header(FILE *f)
{
  byte buf[1024], buf2[1024];
  byte *k, *l;
  uns maxlin = 0;

  buf2[0] = 0;
  l = NULL;
  for(;;)
    {
      if (maxlin++ > max_hdr_lines)
	invalid("0113 Maximal number of header lines (%d) exceeded", max_hdr_lines);
      if (!fgets(buf, sizeof(buf), f))
	retry("0109 Unexpected close while scanning %sheader", additional_hdr ? "additional " : "");
      k = strchr(buf, '\r');
      if (!k)
	k = strchr(buf, '\n');
      if (!k)
	retry("0110 Header line too long");
      *k = 0;
      if (!buf[0])
	break;
      if (buf[0] == ' ' || buf[0] == '\t')
	{
	  if (!l)
	    retry("0112 Invalid continuation line");
	  k = buf;
	  while (*k == ' ' || *k == '\t')
	    k++;
	  *l++ = ' ';
	  strcpy(l, k);
	  l += strlen(k);
	  *l = 0;
	}
      else
	{
	  hdrflush(buf2);
	  strcpy(buf2, buf);
	  l = buf2 + strlen(buf2);
	}
    }
  hdrflush(buf2);
}

static byte *
findhdr(byte *hdr)
{
  struct hdr *h = firsthdr;

  while (h)
    {
      if (!strcasecmp(h->name, hdr))
	return h->contents;
      h = h->next;
    }
  return NULL;
}

/* Data type filtering */

static struct ct_filter *ctfilt;

static byte *
filter_add(struct cfitem *c, byte *l)
{
  return add_ct_filter(ctfilt, l);
}

static inline int
validate_type(byte *t)
{
  return match_ct_filter(ctfilt, t);
}

/* Parsing of response header */

static void
parse_resphdr(void)
{
  byte *t, *l;
  uns code;

  if (!resphdr)
    invalid("0114 Missing response header");
  t = resphdr->name;
  if (strncmp(t, "HTTP/1.", 7))
    invalid("0115 Invalid HTTP version: %s", t);
  t = strchr(t, ' ');
  if (!t)
    goto invreph;
  t++;
  if (!Cdigit(t[0]) || !Cdigit(t[1]) || !Cdigit(t[2]) || (t[3] != ' ' && t[3]))
    goto invreph;
  code = (t[0] - '0')*100 + (t[1] - '0')*10 + t[2] - '0';
  t += 3;
  if (t[0] == ' ')
    t++;
  switch (code)
    {
    case 304:				/* Not modified */
      printf("01%03d %s\n", code, t);
      exit(0);
    case 503:				/* Temporarily Unavailable */
      if (l = findhdr("Retry-After:"))
	printf("r%d\n", (int) atol(l));
      /* Fall-thru */
    case 504:				/* Gateway Timeout */
      printf("?1%03d %s\n", code, t);
      exit(1);
    case 505:				/* HTTP version not supported */
      printf("v1%03d %s\n", code, t);
      exit(1);
    }

  if (code >= 200 && code < 300)	/* Success */
    return;
  if (code >= 300 && code < 400 && code != 305)	/* Redirect (305 == Use Proxy) */
    {
      if (l = findhdr("Location:"))
	printf("R%s\n", l);
      printf("X1%03d %s\n", code, t);
      exit(0);
    }
  invalid("1%03d %s", code, t);

invreph:
  invalid("0116 Invalid response header");
}

/* Parsing and analysis of all incoming headers */

static void
parse_hdr(void)
{
  byte *l;
  byte *base = NULL;

  if (l = findhdr("Content-Location:"))
    base = l;
  if (l = findhdr("Content-Base:"))
    base = l;
  if (base)
    printf("B%s\n", base);
  if (l = findhdr("Content-Encoding:"))
    printf("E%s\n", l);
  if (l = findhdr("Content-Language:"))
    printf("l%s\n", l);
  if (l = findhdr("Content-Type:"))
    {
      printf("T%s\n", l);
      if (!validate_type(l))
	invalid("0120 Content type filtered out");
    }
  if (l = findhdr("Content-Length:"))
    expected_length = atol(l);
  if (l = findhdr("Expires:"))
    {
      uns t = http_parse_date(l);
      if (t)
	{
#if 0
	  if (t < (uns) time(NULL))
	    invalid("0117 Already expired");
#endif
	  printf("e%u\n", t);
	}
    }
  if (l = findhdr("Last-Modified:"))
    {
      uns t = http_parse_date(l);
      if (t)
	{
	  if (if_modified_since && if_modified_since > t)
	    {
	      printf("00118 Not modified\n");
	      exit(0);
	    }
	  printf("L%u\n", t);
	}
    }
  if (l = findhdr("Server:"))
    printf("S%s\n", l);
  if (l = findhdr("Warning:"))
    printf(".%s\n", l);
  if (l = findhdr("Via:"))
    printf("P%s\n", l);
  if (l = findhdr("Transfer-Encoding:"))
    {
      if (!strcasecmp(l, "chunked"))
	is_chunked = 1;
      else
	invalid("0121 Invalid Transfer-Encoding: %s", l);
    }
}

/* Copying of message body */

static void
recv_body_normal(FILE *f, int fh)
{
  byte buf[16384];
  int len;
  uns sum;

  sum = 0;
  while (len = fread(buf, 1, sizeof(buf), f))
    {
      sum += len;
      if (sum > max_body_bytes)
	{
	  printf(".Truncated to %d\n", sum - len);
	  printf("s>%u\n", max_body_bytes);
	  return;
	}
      if (write(fh, buf, len) != len)
	die("Write error: %m");
    }
  if (expected_length && expected_length != sum)
    if (length_checks)
      retry("0118 Length mismatch (got %d, expected %d)", sum, expected_length);
    else
      printf(".Length mismatch (got %d, expected %d)\n", sum, expected_length);
  printf("s%u\n", sum);
}

static void
recv_body_chunked(FILE *f, int fh)
{
  byte buf[16384];
  byte *x;
  uns len, blen, tlen;
  int c;

  if (trace)
    log(L_DEBUG "Using chunked encoding");
  tlen = 0;
  while (fgets(buf, 1024, f))
    {
      x = strchr(buf, '\r');
      if (!x || x[1] != '\n')
	break;
      *x = 0;
      len = 0;
      for(x=buf; Cxdigit(*x); x++)
	len = (len << 4) | Cxvalue(*x);
      if (*x && !Cblank(*x) && *x != ';')
	goto err;
      if (!len)
	{
	  additional_hdr = 1;
	  recv_header(f);
	  printf("s%u\n", tlen);
	  return;
	}
      while (len)
	{
	  if (len > sizeof(buf))
	    blen = sizeof(buf);
	  else
	    blen = len;
	  if (fread(buf, 1, blen, f) != blen)
	    goto unex;
	  if ((uns) write(fh, buf, blen) != blen)
	    die("write error: %m");
	  tlen += blen;
	  if (tlen >= max_body_bytes)
	    {
	      printf(".Truncated to %d\n", tlen);
	      printf("s>%u\n", tlen);
	      return;
	    }
	  len -= blen;
	}
      c = fgetc(f);
      if (c == EOF)
	goto unex;
      if (c != '\r')
	goto err;
      c = fgetc(f);
      if (c == EOF)
	goto unex;
      if (c != '\n')
	goto err;
    }
unex:
  retry("0123 Unexpected close in chunked encoding");
  return;

err:
  invalid("0122 Malformed chunked encoding");
}

static void
recv_body(FILE *f)
{
  struct tempfile t;
  int fh;

  if (dump_to_fd >= 0)
    fh = dump_to_fd;
  else
    {
      open_temp(&t, TF_QUEUE_DATA);
      fh = t.fh;
      printf("@%s\n", t.name);
    }
  if (is_chunked)
    recv_body_chunked(f, fh);
  else
    recv_body_normal(f, fh);
  alarm(0);
  puts("+Downloaded OK");
}

/* Download an object */

static void
download_url(byte *host, uns port, byte *rest)
{
  int sock;
  FILE *f;

  signal(SIGALRM, timeout);

  sock = http_connect(host, port);
  f = fdopen(sock, "w+");
  if (!f)
    die("fdopen failed (%m)");

  alarm(header_timeout);
  send_header(f, host, port, rest);
  if (!use_old_http)
    {
      recv_header(f);
      parse_resphdr();
      parse_hdr();
    }

  alarm(body_timeout);
  recv_body(f);
}

static void
get_url(byte *url)
{
  struct url u;
  byte buf1[MAX_URL_SIZE], buf2[MAX_URL_SIZE], buf3[MAX_URL_SIZE];
  int err;

  if (err = url_canon_split(url, buf1, buf2, &u))
    goto er;
  if (u.protoid != URL_PROTO_HTTP)
    invalid("0100 Invalid protocol");
  if (u.user)
    invalid("0101 Authentication not supported");
  if (err = url_pack(&u, buf3))
    goto er;
  printf("U%s\n", buf3);
  printf("D%u\n", (int) time(NULL));
  if (err = url_enescape(u.rest, buf3))
    goto er;
  download_url(u.host, u.port, buf3);
  return;

er:
  invalid("00%02d %s", err, url_error(err));
}

/* My Arguments, Ignore Null's :-) */

int
main(int argc, char **argv)
{
  initlog(argv[0]);
  if (argc > 4 || argc < 2)
    die("Invalid arguments. Usage: httpget <url> [<if-mod-since> [<use-old-version-flag>]]");
  if (argc >= 3)
    if_modified_since = atol(argv[2]);
  if (argc >= 4)
    use_old_http = atol(argv[3]);
  ctfilt = new_ct_filter();
  cf_read("cf/httpget", cfitems);
  get_url(argv[1]);
  return 0;
}
