/*
 *	Make SMUG Filesystem 1.0
 *
 *	(c) 1998 Martin Mares <mj@ucw.cz>
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <linux/types.h>
#include <linux/fs.h>

#include "smugfs.h"

static void die(char *, ...) __attribute__((noreturn));

static int fd;

static void
die(char *msg, ...)
{
	va_list args;

	va_start(args, msg);
	fprintf(stderr, "mksmugfs: ");
	vfprintf(stderr, msg, args);
	fputc('\n', stderr);
	exit(1);
}

static unsigned char buf[1024];
static unsigned int pos;

static void
write_block(unsigned int i)
{
	if (pos != i && llseek(fd, (((loff_t) i) << 10), SEEK_SET) < 0)
		die("llseek: %m");
	if (write(fd, buf, sizeof(buf)) < 0)
		die("write: %m");
	pos = i+1;
}

int
main(int argc, char **argv)
{
	int size;
	struct stat st;
	struct smug_super_block *sb = (void *) buf;
	struct smug_inode *ii = (void *) buf;
	int i, first_grp, grp_size, num_inodes;

	if (argc != 2 && argc != 3) {
		fprintf(stderr, "Usage: mksmugfs <device> [<blocks>]\n");
		return 1;
	}
	fd = open(argv[1], O_RDWR);
	if (fd < 0)
		die("Unable to open %s: %m", argv[1]);
	if (fstat(fd, &st) < 0)
		die("stat: %m");
	size = (argc > 2) ? atoi(argv[2]) : -1;
	if (S_ISREG(st.st_mode)) {
		if (size < 0)
			size = st.st_size / 1024;
		printf("Creating filesystem in file, size=%d\n", size);
	} else if (!S_ISBLK(st.st_mode))
		die("Must be a file or a block device");
	else {
		if (size < 0) {
			if (ioctl(fd, BLKGETSIZE, &size) > 0)
				printf("BLKGETSIZE returned size %d\n", size);
			else
				die("Device size unknown");
		}
		printf("Creating filesystem on block device, size=%d\n", size);
	}
	if (size < 512)
		die("Filesystem too small");

	puts("Writing super-block...");
	bzero(sb, sizeof(*sb));
	sb->magic = SMUG_SUPER_MAGIC;
	sb->version = SMUG_VERSION;
	sb->flags = SMUG_FLAG_CLEAN;
	sb->total_blocks = size;
	sb->num_dir_blocks = num_inodes = 256;
	sb->first_dir_block = 1;
	sb->first_group_block = first_grp = 1 + sb->num_dir_blocks;
	sb->num_groups = 255;
	grp_size = (sb->total_blocks - first_grp) / sb->num_groups;
	printf("%d groups of %d sectors, %d directory entries\n",
	       sb->num_groups, grp_size, sb->num_dir_blocks);
	write_block(0);

	puts("Writing directory...");
	bzero(ii, sizeof(*ii));
	ii->magic = SMUG_INO_MAGIC;
	ii->type = SMUG_INO_TYPE_ROOT;
	ii->size = 1024;
	ii->mode = S_IFDIR | S_IRWXU | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH;
	ii->mtime = ii->ctime = time(NULL);
	write_block(1);
	bzero(ii, sizeof(*ii));
	ii->magic = SMUG_INO_MAGIC;
	ii->type = SMUG_INO_TYPE_FREE;
	for(i=1; i<num_inodes; i++)
		write_block(i+1);

	puts("Flushing buffers...");
	ioctl(fd, BLKFLSBUF, 0);

	puts("Done.");
	return 0;
}
