/*
 *	Simple MUlti-Gigabyte FileSystem 1.0
 *
 *	(c) 1998 Martin Mares <mj@ucw.cz>
 */

#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/module.h>
#include <linux/malloc.h>
#include <linux/fs.h>
#include <linux/locks.h>
#include <linux/string.h>

#include "smugfs.h"

#define DEBUG

#undef KERN_DEBUG
#define KERN_DEBUG

#ifdef DEBUG
#define DBG(x, y...) printk(KERN_DEBUG "smugfs: " x, ##y)
#define DBG2(x, y...) printk(x, ##y)
#else
#define DBG(x, y...)
#define DBG2(x, y...)
#endif

static struct inode_operations smug_file_inode_ops, smug_dir_inode_ops;

static int
count_bits(void *x, int max)
{
	int i,j;

	j = 0;
	for(i=0; i<max; i++)
		if (test_bit(i, x))
			j++;
	return j;
}

static unsigned int
internal_bmap(struct inode *i, unsigned int sect)
{
	struct smug_super *s = i->i_sb->u.generic_sbp;
	struct buffer_head *h = i->u.generic_ip;
	struct smug_inode *g = (void *) h->b_data;
	unsigned int group, offset;

	group = g->block_map[sect / s->blocks_per_group];
	if (!group) {
		printk(KERN_ERR "smugfs: internal_bmap past file size!\n");
		return s->total_blocks - 1;
	}
	group--;
	offset = sect % s->blocks_per_group;
	return s->first_group_block + s->blocks_per_group * group + offset;
}

static int
smug_bmap(struct inode *i, int blk)
{
	DBG("bmap(%ld,%d)\n", i->i_ino, blk);
	if (blk >= ((i->i_size + BLOCK_SIZE - 1) >> BLOCK_SIZE_BITS))
		return 0;
	else
		return internal_bmap(i, blk);
}

static void
smug_truncate(struct inode *i)
{
	unsigned int low;
	struct smug_super *s = i->i_sb->u.generic_sbp;
	struct buffer_head *h = i->u.generic_ip;
	struct smug_inode *g = (void *) h->b_data;

	DBG("truncate(%ld,%d)\n", i->i_ino, (unsigned int) i->i_size);
	low = (i->i_size + BLOCK_SIZE - 1) >> BLOCK_SIZE_BITS;
	low = (low + s->blocks_per_group - 1) / s->blocks_per_group;
	while (low < 255 && g->block_map[low]) {
		int gr = g->block_map[low] - 1;
		if (!clear_bit(gr, s->group_bitmap))
			printk(KERN_ERR "smugfs: Group %d already free\n", gr);
		g->block_map[low] = 0;
		low++;
	}

	i->i_mtime = i->i_ctime = CURRENT_TIME;
	i->i_dirt = 1;
}

static int
smug_read(struct inode *i, struct file *f, char *buf, int size)
{
	unsigned int size0, res;
	loff_t pos = f->f_pos;

	DBG("read(%ld,%d)\n", i->i_ino, size);
	res = -EIO;
	if (pos + size > i->i_size) {
		size = i->i_size - pos;
		if (size <= 0)
			return 0;
	}

	size0 = size;
	while (size) {
		unsigned int sect = pos >> BLOCK_SIZE_BITS;
		unsigned int rel = pos & (BLOCK_SIZE - 1);
		unsigned int howmuch;
		struct buffer_head *h;

		howmuch = BLOCK_SIZE - rel;
		if (howmuch > size)
			howmuch = size;
		sect = internal_bmap(i, sect);
		DBG("slow_read(%d,%d,%d)\n", sect, rel, howmuch);
		h = breada(i->i_sb->s_dev, sect, BLOCK_SIZE, 0, (size < 65536) ? size : 65536);
		if (!h)
			goto done;
		memcpy_tofs(buf, h->b_data + rel, howmuch);
		brelse(h);
		size -= howmuch;
		pos += howmuch;
		buf += howmuch;
	}

	res = size0;

done:
	if (res >= 0)
		f->f_pos = pos;
	return res;
}

static int
slow_write(struct inode *i, const char *buf, int size, int clear, loff_t where)
{
	unsigned int sect = where >> BLOCK_SIZE_BITS;
	unsigned int pos = where & (BLOCK_SIZE - 1);
	struct buffer_head *h;

	sect = internal_bmap(i, sect);
	DBG("slow_write(%d,%d,%d,%d)\n", sect, pos, size, clear);
	if (!clear)
		h = bread(i->i_sb->s_dev, sect, BLOCK_SIZE);
	else {
		h = getblk(i->i_sb->s_dev, sect, BLOCK_SIZE);
		if (h) {
			if (size != BLOCK_SIZE)
				memset(h->b_data, 0, BLOCK_SIZE);
			mark_buffer_uptodate(h, 1);
		}
	}
	if (!h) {
		smug_truncate(i);
		return 1;
	}
	memcpy_fromfs(h->b_data + pos, buf, size);
	update_vm_cache(i, where, h->b_data + pos, size);
	mark_buffer_dirty(h, 0);
	brelse(h);
	return 0;
}

static int
smug_write(struct inode *i, struct file *f, const char *buf, int size)
{
	unsigned int low, high, sect, rem, size0, res;
	loff_t pos = f->f_pos;
	struct smug_super *s = i->i_sb->u.generic_sbp;
	struct buffer_head *h = i->u.generic_ip;
	struct smug_inode *g = (void *) h->b_data;

	DBG("write(%ld,%d)\n", i->i_ino, size);
	res = -EIO;
	if (pos + size > i->i_size) {
		low = (i->i_size + BLOCK_SIZE - 1) >> BLOCK_SIZE_BITS;		/* First block after file */
		low = (low + s->blocks_per_group - 1) / s->blocks_per_group;	/* First group after file */
		high = (pos + size - 1) >> BLOCK_SIZE_BITS;			/* Last block of new file */
		high = high / s->blocks_per_group;				/* Last group of new file */
		if (high > 255)
			return -E2BIG;
		while (low <= high) {
			if (g->block_map[low]) {
				printk(KERN_ERR "smugfs: Block already allocated!\n");
				return -EIO;
			}
			sect = find_first_zero_bit(s->group_bitmap, s->num_groups);
			if (sect >= s->num_groups) {
				smug_truncate(i);
				return -ENOSPC;
			}
			set_bit(sect, s->group_bitmap);
			g->block_map[low++] = sect+1;
			DBG("Allocated group %d\n", sect);
		}
	}

	rem = BLOCK_SIZE - (pos & (BLOCK_SIZE - 1));
	size0 = size;
	if (rem != BLOCK_SIZE) {
		if (rem > size)
			rem = size;
		if (slow_write(i, buf, rem, 0, pos))
			goto done;
		size -= rem;
		buf += rem;
		pos += rem;
	}

	while (size >= BLOCK_SIZE) {
		if (slow_write(i, buf, BLOCK_SIZE, 1, pos))
			goto done;
		size -= BLOCK_SIZE;
		buf += BLOCK_SIZE;
		pos += BLOCK_SIZE;
	}

	if (size) {
		if (slow_write(i, buf, size, (pos >= i->i_size), pos))
			goto done;
		pos += size;
	}

	res = size0;

done:
	if (res >= 0) {
		f->f_pos = pos;
		if (pos > i->i_size)
			i->i_size = pos;
	}
	i->i_mtime = CURRENT_TIME;
	i->i_dirt = 1;
	return res;
}

static int
smug_dir_read(struct inode *i, struct file *f, char *buf, int size)
{
	DBG("dir_read\n");
	return -EISDIR;
}

static int
smug_readdir(struct inode *i, struct file *f, void *buf, filldir_t func)
{
	int pos = f->f_pos;
	struct smug_super *s;

	DBG("readdir(%d)\n", pos);
	if (!i || !i->i_sb || !S_ISDIR(i->i_mode))
		return -EBADF;
	s = i->i_sb->u.generic_sbp;
	if (pos < 0)
		pos = 0;
	if (!pos) {
		if (func(buf, ".", 1, pos, SMUG_ROOT_INODE))
			return 0;
		f->f_pos = ++pos;
	}
	if (pos == 1) {
		if (func(buf, "..", 2, pos, SMUG_ROOT_INODE))
			return 0;
		f->f_pos = ++pos;
	}
	pos--;
	while (pos < s->num_dir_blocks) {
		if (test_bit(pos, s->valid_dir_bitmap)) {
			struct inode *d;
			struct buffer_head *bh;
			struct smug_inode *e;
			if (!(d = iget(i->i_sb, pos)))
				return -EIO;
			if (!(bh = d->u.generic_ip)) {
				iput(d);
				return -EIO;
			}
			e = (void *) bh->b_data;
			if (func(buf, e->name, strlen(e->name), pos, d->i_ino)) {
				iput(d);
				return 0;
			}
			iput(d);
		}
		f->f_pos = ++pos + 1;
	}
	return 0;
}

static struct inode *
smug_find_dir_entry(struct inode *i, const char *name, int namelen)
{
	struct super_block *sb = i->i_sb;
	struct smug_super *s = sb->u.generic_sbp;
	int pos;
	struct inode *j;

	for (pos=1; pos<s->num_dir_blocks; pos++)
		if (test_bit(pos, s->valid_dir_bitmap) && (j = iget(sb, pos))) {
			struct buffer_head *h = j->u.generic_ip;
			if (h) {
				struct smug_inode *d = (struct smug_inode *) h->b_data;
				if (!strcmp(name, d->name))
					return j;
			}
			iput(j);
		}
	return NULL;
}

static int
smug_create(struct inode *i, const char *name, int namelen, int mode, struct inode **result)
{
	struct super_block *sb = i->i_sb;
	struct smug_super *s = sb->u.generic_sbp;
	int pos;
	struct inode *j;
	struct buffer_head *bh;
	struct smug_inode *c;

	DBG("create(%s)\n", name);
	if (namelen > 255) {
		iput(i);
		return -ENAMETOOLONG;
	}
	pos = find_first_zero_bit(s->dir_bitmap, s->num_dir_blocks);
	if (pos >= s->num_dir_blocks) {
		iput(i);
		return -ENOSPC;
	}
	set_bit(pos, s->dir_bitmap);
	bh = getblk(sb->s_dev, s->first_dir_block+pos, BLOCK_SIZE);
	if (! (j = get_empty_inode())) {
		clear_bit(pos, s->dir_bitmap);
		brelse(bh);
		iput(i);
		return -ENOMEM;
	}
	c = (struct smug_inode *) bh->b_data;
	memset(c, 0, sizeof(*c));
	c->magic = SMUG_INO_MAGIC;
	c->type = SMUG_INO_TYPE_FILE;
	memcpy(c->name, name, namelen);
	j->i_dev = sb->s_dev;
	j->i_ino = pos;
	j->i_mode = mode;
	j->i_uid = current->fsuid;
	j->i_gid = current->fsgid;
	j->i_size = 0;
	j->i_mtime = j->i_ctime = j->i_atime = CURRENT_TIME;
	j->i_blksize = 0;
	j->i_blocks = 0;
	j->i_op = &smug_file_inode_ops;
	j->i_sb = sb;
	j->i_flags = sb->s_flags;
	j->i_dirt = 1;
	j->u.generic_ip = bh;
	i->i_mtime = CURRENT_TIME;
	i->i_dirt = 1;
	i->i_version = ++event;
	mark_buffer_uptodate(bh, 1);
	insert_inode_hash(j);
	iput(i);
	set_bit(pos, s->valid_dir_bitmap);
	*result = j;
	return 0;
}

static int
smug_lookup(struct inode *i, const char *name, int namelen, struct inode **result)
{
	struct inode *j;

	DBG("lookup(%s)\n", name);
	if (namelen == 1 && name[0] == '.') {
		*result = i;
		return 0;
	}
	j = smug_find_dir_entry(i, name, namelen);
	if (!j) {
		iput(i);
		return -ENOENT;
	}
	iput(i);
	*result = j;
	return 0;
}

static int
smug_unlink(struct inode *i, const char *name, int namelen)
{
	struct inode *j;
	int ret;

	DBG("unlink(%s)\n", name);
	j = smug_find_dir_entry(i, name, namelen);
	if (!j) {
		iput(i);
		return -ENOENT;
	}
	ret = -EPERM;
	if (S_ISDIR(j->i_mode))
		goto done;
	if ((i->i_mode & S_ISVTX) &&
	    current->fsuid != i->i_uid &&
	    current->fsuid != j->i_uid &&
	    !fsuser())
		goto done;
	if (!j->i_nlink)
		printk(KERN_ERR "smugfs: Deleting ghost file\n");
	else
		j->i_nlink--;
	i->i_version = ++event;
	j->i_ctime = i->i_ctime = i->i_mtime = CURRENT_TIME;
	i->i_dirt = 1;
	j->i_dirt = 1;
	ret = 0;
done:
	iput(i);
	iput(j);
	return ret;
}

static int
smug_rename(struct inode *oi, const char *oname, int olen, struct inode *ni, const char *nname, int nlen, int must_be_dir)
{
	struct inode *j, *k;
	int result = 0;

	DBG("rename(%s,%s)\n", oname, nname);
	if (olen >= 255 || nlen >= 255)
		result = -ENAMETOOLONG;
	else if (! (j = smug_find_dir_entry(oi, oname, olen)))
		result = -ENOENT;
	else {
		struct buffer_head *bh = j->u.generic_ip;
		if (must_be_dir && !S_ISDIR(j->i_mode))
			result = -ENOENT;
		else if (!bh)
			result = -EIO;
		else {
			struct smug_inode *ii = (void *) bh->b_data;
			k = smug_find_dir_entry(ni, nname, nlen);
			if (k) {
				if (!S_ISDIR(k->i_mode)) {
					if ((oi->i_mode & S_ISVTX) &&
					    current->fsuid != oi->i_uid &&
					    current->fsuid != j->i_uid &&
					    !fsuser())
						result = -EPERM;
					else {
						k->i_nlink--;
						k->i_dirt = 1;
					}
				}
				iput(k);
			}
			if (!result) {
				strcpy(ii->name, nname);
				j->i_dirt = 1;
			}
		iput(j);
		}
	}
	iput(oi);
	iput(ni);
	return result;
}

static struct file_operations smug_file_fops = {
	NULL,						    /* lseek */
	smug_read,					    /* Use generic_file_read if you have 64-bit page-cache */
	smug_write,
	NULL,						    /* readdir */
	NULL,						    /* select */
	NULL,						    /* ioctl */
	generic_file_mmap,
	NULL,						    /* open */
	NULL,						    /* release */
	NULL,						    /* fsync */
	NULL,						    /* fasync */
	NULL,						    /* check_media_type */
	NULL						    /* revalidate */
};

static struct inode_operations smug_file_inode_ops = {
	&smug_file_fops,
	NULL,						    /* create */
	NULL,						    /* lookup */
	NULL,						    /* link */
	NULL,						    /* unlink */
	NULL,						    /* symlink */
	NULL,						    /* mkdir */
	NULL,						    /* rmdir */
	NULL,						    /* mknod */
	NULL,						    /* rename */
	NULL,						    /* readlink */
	NULL,						    /* follow_link */
	generic_readpage,
	NULL,						    /* writepage */
	smug_bmap,
	smug_truncate,
	NULL,						    /* permission */
	NULL						    /* smap */
};

static struct file_operations smug_dir_fops = {
	NULL,						    /* lseek */
	smug_dir_read,
	NULL,
	smug_readdir,
	NULL,						    /* select */
	NULL,						    /* ioctl */
	NULL,						    /* mmap */
	NULL,						    /* open */
	NULL,						    /* release */
	NULL,						    /* fsync */
	NULL,						    /* fasync */
	NULL,						    /* check_media_type */
	NULL						    /* revalidate */
};

static struct inode_operations smug_dir_inode_ops = {
	&smug_dir_fops,
	smug_create,
	smug_lookup,
	NULL,						    /* link */
	smug_unlink,
	NULL,						    /* symlink */
	NULL,						    /* mkdir */
	NULL,						    /* rmdir */
	NULL,						    /* mknod */
	smug_rename,
	NULL,						    /* readlink */
	NULL,						    /* follow_link */
	NULL,						    /* readpage */
	NULL,						    /* writepage */
	NULL,						    /* bmap */
	NULL,						    /* truncate */
	NULL,						    /* permission */
	NULL						    /* smap */
};

static void
smug_read_inode(struct inode *i)
{
	struct buffer_head *bh;
	struct smug_inode *ii;
	struct smug_super *s;

	DBG("read_inode(%ld)\n", i->i_ino);

#ifdef DEBUG
	if (i->u.generic_ip)
		printk(KERN_ERR "smugfs: read_inode on already read inode\n");
#endif

	s = i->i_sb->u.generic_sbp;
	if (i->i_ino >= s->num_dir_blocks) {
		printk(KERN_ERR "smugfs: Requested inode %ld is out of range\n", i->i_ino);
	bad:
		make_bad_inode(i);
		return;
	}
	bh = bread(i->i_sb->s_dev, s->first_dir_block + i->i_ino, BLOCK_SIZE);
	if (!bh) {
		printk(KERN_ERR "smugfs: Cannot read inode %ld\n", i->i_ino);
		goto bad;
	}

	ii = (struct smug_inode *) bh->b_data;
	if (ii->magic != SMUG_INO_MAGIC || ii->type > SMUG_INO_TYPE_ROOT) {
		printk(KERN_ERR "smugfs: Invalid inode %ld\n", i->i_ino);
	bad2:
		brelse(bh);
		goto bad;
	}

	if (S_ISREG(ii->mode) && ii->type == SMUG_INO_TYPE_FILE)
		i->i_op = &smug_file_inode_ops;
	else if (S_ISDIR(ii->mode) && ii->type == SMUG_INO_TYPE_ROOT)
		i->i_op = &smug_dir_inode_ops;
	else {
		printk(KERN_ERR "smugfs: Invalid inode[%ld] type\n", i->i_ino);
		goto bad2;
	}

	i->u.generic_ip = bh;
	i->i_mode = ii->mode;
	i->i_nlink = 1;
	i->i_uid = ii->uid;
	i->i_gid = ii->gid;
	i->i_size = ii->size;
	i->i_atime = ii->mtime;
	i->i_ctime = ii->ctime;
	i->i_mtime = ii->mtime;
	i->i_blksize = PAGE_SIZE;
	i->i_blocks = 0;
}

static void
smug_write_inode(struct inode *i)
{
	struct buffer_head *bh = i->u.generic_ip;
	struct smug_inode *ii;

	DBG("write_inode(%ld)\n", i->i_ino);
	if (!bh) {
		DBG("write_inode on invalid inode\n");
		return;
	}

	ii = (struct smug_inode *) bh->b_data;

	ii->size = i->i_size;
	ii->uid = i->i_uid;
	ii->gid = i->i_gid;
	ii->mode = i->i_mode;
	ii->mtime = i->i_mtime;
	ii->ctime = i->i_ctime;

	mark_buffer_dirty(bh, 1);
	i->i_dirt = 0;
}

static void
smug_put_inode(struct inode *i)
{
	DBG("put_inode(%ld)\n", i->i_ino);
	if (!i->i_nlink) {
		struct buffer_head *bh = i->u.generic_ip;
		if (bh) {
			struct smug_inode *si = (void *) bh->b_data;
			struct super_block *sb = i->i_sb;
			struct smug_super *s = sb->u.generic_sbp;
			DBG("Deleting it...\n");
			clear_bit(i->i_ino, s->valid_dir_bitmap);
			clear_bit(i->i_ino, s->dir_bitmap);
			i->i_size = 0;
			si->type = SMUG_INO_TYPE_FREE;
			smug_truncate(i);
			smug_write_inode(i);
			clear_inode(i);
		}
	}
}

static void
smug_clear_inode(struct inode *i)
{
	struct buffer_head *bh = i->u.generic_ip;

	DBG("clear_inode(%ld)\n", i->i_ino);
	if (bh)
		brelse(bh);
}

static void
smug_put_super(struct super_block *sb)
{
	struct smug_super *s = sb->u.generic_sbp;
	struct buffer_head *bh = s->super_bh;
	struct smug_super_block *ss = (struct smug_super_block *) bh->b_data;

	DBG("put_super\n");
	lock_super(sb);
	if (s->old_flags & SMUG_FLAG_CLEAN) {
		ss->flags |= SMUG_FLAG_CLEAN;
		mark_buffer_dirty(bh, 1);
	}
	brelse(bh);
	kfree_s(s, sizeof(*s));
	sb->s_dev = 0;
	unlock_super(sb);
	MOD_DEC_USE_COUNT;
	DBG("unmounted\n");
}

static void
smug_statfs(struct super_block *sb, struct statfs *buf, int size)
{
	struct statfs t;
	struct smug_super *s = sb->u.generic_sbp;

	DBG("statfs()\n");
	t.f_type = sb->s_magic;
	t.f_bsize = sb->s_blocksize;
	t.f_blocks = s->total_blocks;
	t.f_bavail = t.f_bfree = s->blocks_per_group * (s->num_groups - count_bits(s->group_bitmap, s->num_groups));
	t.f_files = s->num_dir_blocks;
	t.f_ffree = s->num_dir_blocks - count_bits(s->dir_bitmap, s->num_dir_blocks);
	t.f_namelen = 255;
	memcpy_tofs(buf, &t, size);
}

static int
smug_remount(struct super_block *sb, int *flags, char *data)
{
	struct smug_super *s = sb->u.generic_sbp;
	struct buffer_head *bh = s->super_bh;
	struct smug_super_block *ss = (struct smug_super_block *) bh->b_data;

	DBG("Remounting: old=%08lx, new=%08x\n", sb->s_flags, *flags);
	if (! ((*flags ^ sb->s_flags) & MS_RDONLY))
		return 0;
	lock_super(sb);
	if (*flags & MS_RDONLY) {			    /* Remounting R/O */
		if (s->old_flags & SMUG_FLAG_CLEAN) {
			ss->flags |= SMUG_FLAG_CLEAN;
			mark_buffer_dirty(bh, 1);
		}
	} else {					    /* Remounting R/W */
		if (!(s->old_flags & SMUG_FLAG_CLEAN)) {
			printk(KERN_WARNING "smugfs: Remounting unchecked fs read-write\n");
		} else {
			ss->flags &= ~SMUG_FLAG_CLEAN;
			mark_buffer_dirty(bh, 1);
		}
	}
	unlock_super(sb);
	return 0;
}

static struct super_operations smug_super_ops = {
	smug_read_inode,
	NULL,					/* notify_change */
	smug_write_inode,
	smug_put_inode,
	smug_put_super,
	NULL,					/* write_super */
	smug_statfs,
	smug_remount,
	smug_clear_inode
};

static char *
scan_dir(struct super_block *sb)
{
	struct buffer_head *bh;
	struct smug_inode *ii;
	struct smug_super *s = sb->u.generic_sbp;
	int i, j, cnt;
	unsigned int bg;

	for(i=0; i<s->num_dir_blocks; i++) {
		bh = bread(sb->s_dev, s->first_dir_block + i, BLOCK_SIZE);
		if (!bh)
			return "Inode read error";
		ii = (struct smug_inode *) bh->b_data;
		if (ii->magic != SMUG_INO_MAGIC)
			return "Bad inode magic";
		switch (ii->type) {
		case SMUG_INO_TYPE_FREE:
			break;
		case SMUG_INO_TYPE_ROOT:
			if (i)
				return "Two root nodes";
			set_bit(i, s->dir_bitmap);
			break;
		case SMUG_INO_TYPE_FILE:
			cnt = 0;
			set_bit(i, s->dir_bitmap);
			for(j=0; j<256; j++)
				if (ii->block_map[j]) {
					int grp = ii->block_map[j] - 1;
					if (grp >= s->num_groups)
						return "Block out of range";
					if (set_bit(grp, s->group_bitmap))
						return "Cross-linked files";
					cnt++;
				}
			bg = (ii->size + BLOCK_SIZE - 1) >> BLOCK_SIZE_BITS;
			bg = (bg + s->blocks_per_group - 1) / s->blocks_per_group;
			if (bg != cnt)
				return "Misallocated inode";
			break;
		default:
			return "Invalid inode type";
		}
		brelse(bh);
	}

	memcpy(s->valid_dir_bitmap, s->dir_bitmap, sizeof(s->dir_bitmap));
	return NULL;
}

static struct super_block *
smug_read_super(struct super_block *sb, void *data, int silent)
{
	struct buffer_head *bh;
	struct smug_super_block *ss;
	struct smug_super *s;
	kdev_t dev = sb->s_dev;
	char *msg;

	DBG("read_super\n");
	MOD_INC_USE_COUNT;
	lock_super(sb);
	set_blocksize(dev, BLOCK_SIZE);
	bh = bread(dev, 0, BLOCK_SIZE);
	if (!bh) {
		msg = "Unable to read superblock";
	err:
		if (bh)
		       brelse(bh);
		if (!silent)
			printk(KERN_ERR "smugfs: %s on device %s\n", msg, kdevname(dev));
		unlock_super(sb);
		MOD_DEC_USE_COUNT;
		return NULL;
	}
	ss = (struct smug_super_block *) bh->b_data;
	if (ss->magic != SMUG_SUPER_MAGIC) {
		msg = "Unrecognised superblock magic";
		goto err;
	}
	if (ss->version != SMUG_VERSION) {
		msg = "Invalid fs version";
		goto err;
	}
	if (ss->num_groups > 255 || ss->num_dir_blocks > 256) {
		msg = "Filesystem too large";
		goto err;
	}
	if (!(ss->flags & SMUG_FLAG_CLEAN))
		printk(KERN_WARNING "smugfs: Filesystem not unmounted cleanly, please run fsck.\n");
	if (!(s = kmalloc(sizeof(struct smug_super), GFP_KERNEL))) {
		msg = "Out of memory";
		goto err;
	}
	memset(s, 0, sizeof(*s));
	s->old_flags = ss->flags;
	s->total_blocks = ss->total_blocks;
	s->num_groups = ss->num_groups;
	s->first_group_block = ss->first_group_block;
	s->blocks_per_group = (s->total_blocks - s->first_group_block) / s->num_groups;
	s->first_dir_block = ss->first_dir_block;
	s->num_dir_blocks = ss->num_dir_blocks;
	DBG("%d blocks (%d groups of %d blocks at %d + directory %d blocks at %d + %d blocks trash)\n",
	    s->total_blocks,
	    s->num_groups,
	    s->blocks_per_group,
	    s->first_group_block,
	    s->num_dir_blocks,
	    s->first_dir_block,
	    s->total_blocks - s->num_groups * s->blocks_per_group - s->num_dir_blocks - 1 /* SB */ );

	sb->u.generic_sbp = s;
	sb->s_blocksize = BLOCK_SIZE;
	sb->s_blocksize_bits = BLOCK_SIZE_BITS;
	sb->s_magic = SMUG_SUPER_MAGIC;
	sb->s_op = &smug_super_ops;

	DBG("Scanning directory\n");
	if (msg = scan_dir(sb)) {
		kfree_s(s, sizeof(*s));
		goto err;
	}
	DBG("%d inodes and %d groups used\n",
	    count_bits(s->dir_bitmap, s->num_dir_blocks),
	    count_bits(s->group_bitmap, s->num_groups));

	unlock_super(sb);				    /* Read the root inode */
	if (! (sb->s_mounted = iget(sb, SMUG_ROOT_INODE)) || !sb->s_mounted->u.generic_ip) {
		printk(KERN_ERR "smugfs: Error reading root inode\n");
		brelse(bh);
		kfree_s(s, sizeof(*s));
		MOD_DEC_USE_COUNT;
		return NULL;
	}

	if (ss->flags & SMUG_FLAG_CLEAN) {
		ss->flags &= ~SMUG_FLAG_CLEAN;
		mark_buffer_dirty(bh, 1);
	}

	DBG("Mounted\n");
	s->super_bh = bh;
	return sb;
}

static struct file_system_type smug_fs_type = {
	smug_read_super, "smug", 1, NULL
};

#ifdef MODULE
int init_module(void)
#else
int init_smug_fs(void)
#endif
{
	int stat;

	printk("smugfs 1.0 (c) 1998 Martin Mares <mj@ucw.cz>\n");
	stat = register_filesystem(&smug_fs_type);
	if (stat)
		printk(KERN_ERR "Unable to register!\n");
	return stat;
}

#ifdef MODULE

void
cleanup_module(void)
{
	DBG("Removing\n");
	unregister_filesystem(&smug_fs_type);
}

#endif
