/*
 *	Sherlock Search Engine -- Main Loop
 *
 *	(c) 1997 Martin Mares, <mj@atrey.karlin.mff.cuni.cz>
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <sys/time.h>

#include "sherlock.h"

#include <lib/string.h>

static list query_list;

static void
close_conn(struct query *q, byte *err)
{
  if (err)
    add_reply(&q->rbuf, err);
  flush_reply_buf(q, &q->rbuf);
  reply_string(q, "\n", 1);
  flush_query(q);
  close(q->fd);
  rem_node(&q->n);
  free_query(q);
}

static void
do_request(struct query *q)
{
  byte *err;

  init_timer();
  current_query = q;
  query_rbuf = &q->rbuf;
  cache_rbuf = NULL;

  if (log_requests > 1)
    log(L_INFO "<< %s", q->inbuf);
  if (!q->inbuf[0])
    add_qr("-102 Empty request");
  else if (err = parse_query(q->inbuf))
    {
      if (err[0] == '-')
	add_qr(err);
      else
	add_qr("-102 %s", err);
    }
  else if (q->expr)
    {
      if (q->from >= num_matches)
	add_qr("-106 FROM out of range");
      else
	do_query(q);
    }
  else
    do_command(q);
}

static void
incoming_conn(int sock)
{
  int sock2, alen;
  struct sockaddr_in addr;
  u32 ip;
  struct query *q;

  alen = sizeof(addr);
  sock2 = accept(sock, (struct sockaddr *) &addr, &alen);
  if (sock2 < 0)
    {
      if (errno != EAGAIN)
	log(L_ERROR "accept(): %m");
      return;
    }
  if (fcntl(sock2, F_SETFL, O_NONBLOCK))
    {
      log(L_ERROR "fcntl(): %m");
      return;
    }

  q = new_query();
  q->fd = sock2;
  q->established = time(NULL);
  add_tail(&query_list, &q->n);

  ip = ntohl(addr.sin_addr.s_addr);
  if (!check_access(ip))
    {
      if (log_rejected)
	log(L_INFO "Rejected connection from %08x", ip);
      close_conn(q, "-100 Incoming connection rejected due to insufficient voltage. You are not allowed to talk to me.");
      return;
    }

  if (log_incoming)
    log(L_INFO "Accepted connection from %08x, fd=%d", ip, sock2);
}

static void
continue_conn(struct query *q)
{
  int sz, mx;
  byte *r;

  mx = q->ibend - q->ibptr - 1;
  sz = read(q->fd, q->ibptr, mx);
  DBG("Data from %d: %d bytes", q->fd, sz);
  if (!sz)
    {
      close_conn(q, "-108 Incomplete request");
      return;
    }
  if (sz < 0)
    {
      if (errno == EAGAIN)
	return;
      close_conn(q, "-108 Read error");
      return;
    }
  if (sz >= mx)
    {
      close_conn(q, "-101 Request too long");
      return;
    }
  r = q->ibptr;
  while (sz--)
    {
      if (*r == '\r' || *r == '\n')
	{
	  while (r > q->inbuf && Cspace(r[-1] == '\t'))
	    r--;
	  *r = 0;
	  do_request(q);
	  close_conn(q, NULL);
	  return;
	}
      r++;
    }
  q->ibptr = r;
}

static void
mainloop(void)
{
  int sock, n;
  int one = 1;
  time_t now;
  fd_set fds;
  struct query *q, *nq;
  struct timeval tv;
  struct sockaddr_in addr;
  uns timeout;

  signal(SIGPIPE, SIG_IGN);
  sock = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
  if (sock < 0)
    die("socket(): %m");
  if (setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &one, sizeof(one)))
    die("setsockopt(): %m");
  if (fcntl(sock, F_SETFL, O_NONBLOCK))
    die("fcntl(): %m");

  addr.sin_family = AF_INET;
  addr.sin_port = htons(port);
  addr.sin_addr.s_addr = INADDR_ANY;
  if (bind(sock, (struct sockaddr *) &addr, sizeof(addr)))
    die("bind(): %m");
  if (listen(sock, listen_queue))
    die("listen(): %m");
  log(L_INFO "Listening at port %d", port);

  init_list(&query_list);

  for(;;)
    {
      now = time(NULL);
      n = sock;
      FD_ZERO(&fds);
      FD_SET(sock, &fds);
      timeout = 1000000;
      for(q = HEAD(query_list); nq = (void *) q->n.next; q = nq)
	{
	  if (now - q->established >= connection_timeout)
	    close_conn(q, "-107 Timed out");
	  else
	    {
	      uns rem = q->established + connection_timeout - now;
	      if (rem < timeout)
		timeout = rem;
	      FD_SET(q->fd, &fds);
	      if (q->fd > n)
		n = q->fd;
	    }
	}

      tv.tv_sec = timeout;
      tv.tv_usec = 0;
      n = select(n+1, &fds, NULL, NULL, &tv);
      if (n < 0)
	die("Select failed: %m");
      if (!n)
	continue;

      if (FD_ISSET(sock, &fds))
	incoming_conn(sock);

      for(q = HEAD(query_list); nq = (void *) q->n.next; q = nq)
	if (FD_ISSET(q->fd, &fds))
	  continue_conn(q);
    }
 }

int
main(int argc, char **argv)
{
  initlog(argv[0]);
#ifdef DEBUG
  if (argc == 2)
    {
      global_debug = strtol(argv[1], NULL, 0);
      argc--;
    }
#endif
  if (argc != 1)
    die("This program has no parameters");
  read_config();
  log(L_INFO "Starting...");
  cache_init();
  file_cache_init();
  word_init();
  nonword_init();
  ref_init();
  objects_init();
  pseudo_init();
  mainloop();
  return 0;
}
