/*
 *	Sherlock Search Engine -- Pseudo-Objects
 *
 *	(c) 1997--1999 Martin Mares <mj@atrey.karlin.mff.cuni.cz>
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/mman.h>

#include "sherlock.h"

#include <lib/pools.h>
#include <lib/string.h>
#include <lib/searchdb.h>
#include <charset/unicode.h>

#define MAX_PWORD_LEN 64

struct pword {
  struct pword *next;
  int refcnt;
  uns refnum;
  struct sdbitem *refs;
  char text[1];
};

static struct pword **pword_hash, **pword_index;
static uns num_pseudos, pword_hash_cnt, pseudos_size, pword_count;
static int pseudo_fd;
static struct mempool *pseudo_pool;
static byte *pseudos, **pseudo_ptrs;

static void
pseudo_free(void)
{
  if (pseudo_pool)
    {
      free_pool(pseudo_pool);
      pseudo_pool = NULL;
    }
  if (pseudo_fd > 0)
    {
      close(pseudo_fd);
      pseudo_fd = 0;
    }
  if (pseudos)
    {
      munmap(pseudos, pseudos_size);
      pseudos = NULL;
    }
  num_pseudos = pseudos_size = pword_count = pword_hash_cnt = 0;
}

static void
pseudo_objs(void)
{
  byte *i = pseudos;
  byte **p;

  num_pseudos = 0;
  while (*i)
    {
      while (*i == '\n')
	i++;
      num_pseudos++;
      while (*i && (*i != '\n' || i[1] != '\n'))
	i++;
    }
  DBG("Found %d pseudos", num_pseudos);
  p = pseudo_ptrs = pool_alloc(pseudo_pool, sizeof(byte *) * (num_pseudos + 1));
  i = pseudos;
  while (*i)
    {
      while (*i == '\n')
	i++;
      *p++ = i;
      while (*i && (*i != '\n' || i[1] != '\n'))
	i++;
    }
  *p = i;
}

static uns
pword_canon(byte *pw, word *can)
{
  uns u;
  int len = 0;

  for(;;)
    {
      GET_UTF8(pw, u);
      if (!u)
	break;
      *can++ = Utoupper(Uunaccent(u));
      len++;
    }
  *can = 0;
  return len;
}

static uns
pword_hash_fn(word *c, uns len)
{
  uns i = len;

  while (*c)
    i = i * 313 + *c++;
  return i % pword_hash_cnt;
}

static struct pword *
pword_find(byte *w, int create)
{
  word cax[MAX_PWORD_LEN], cay[MAX_PWORD_LEN];
  byte new[MAX_PWORD_LEN*4];
  uns lex, ley, h;
  struct pword *p;

  lex = pword_canon(w, cax);
  h = pword_hash_fn(cax, lex);
  p = pword_hash[h];
  DBG("<%s> %d", w, h);
  while (p)
    {
      ley = pword_canon(p->text, cay);
      if (lex == ley && !memcmp(cax, cay, sizeof(word) * lex))
	return p;
      p = p->next;
    }
  if (!create)
    return NULL;
  ley = ucs2_to_utf8(new, cax);
  DBG("New: <%s>", new);
  p = pool_alloc(pseudo_pool, sizeof(struct pword) + ley);
  p->next = pword_hash[h];
  pword_hash[h] = p;
  p->refcnt = 0;
  p->refs = NULL;
  strcpy(p->text, new);
  pword_count++;
  return p;
}

static byte *
pseudo_attr(byte *e, uns a)
{
  for(;;)
    {
      if (Cctrl(*e))
	return NULL;
      if (*e == a)
	return e+1;
      while (*e && *e++ != '\n')
	;
    }
}

static void
do_for_all_pwords(void (*func)(uns o, byte *w))
{
  uns ob;

  for(ob=0; ob<num_pseudos; ob++)
    {
      byte *e = pseudo_ptrs[ob];
      if (!(e = pseudo_attr(e, 'W')))
	return;
      for(;;)
	{
	  byte w[MAX_PWORD_LEN], *f = w;

	  while (Cblank(*e))
	    e++;
	  if (Cctrl(*e))
	    break;
	  while (!Cblank(*e) && !Cctrl(*e))
	    {
	      if (f >= w + sizeof(w) - 1)
		break;
	      *f++ = *e++;
	    }
	  *f = 0;
	  func(ob, w);
	}
    }
}

static void
pword_cnt(uns ob, byte *e)
{
  struct pword *p = pword_find(e, 1);
  p->refcnt++;
}

static void
pword_sdb(uns ob, byte *e)
{
  struct pword *p = pword_find(e, 0);
  struct sdbitem *s;

  if (!p)
    die("pword_sdb: Wort nicht gefunden!");
  if (!p->refs)
    {
      p->refs = pool_alloc(pseudo_pool, p->refcnt * sizeof(struct sdbitem));
      p->refcnt = 0;
      p->refnum = FIRST_PSEUDO_REF + pword_count;
      pword_index[pword_count++] = p;
    }
  s = p->refs + p->refcnt++;
  s->oid = FIRST_PSEUDO_OBJ + ob;
  s->tag = ANY_TAG;
  s->flags = (WT_WORD << WT_SHIFT) | WQ_MASK;
#ifdef SHERLOCK_CONFIG_REF_WEIGHTS
  s->weight = 0;
#endif
}

static void
pseudo_words(void)
{
  uns exp = num_pseudos / 2;
  byte *e;
  uns ob;

  pword_hash_cnt = 16;
  pword_count = 0;
  while (pword_hash_cnt < exp)
    pword_hash_cnt *= 2;
  DBG("Hashing pwords to %d buckets", pword_hash_cnt);
  pword_hash = pool_alloc(pseudo_pool, sizeof(struct pword *) * pword_hash_cnt);
  bzero(pword_hash, sizeof(struct pword *) * pword_hash_cnt);
  do_for_all_pwords(pword_cnt);
  DBG("Total %d words", pword_count);
  pword_index = pool_alloc(pseudo_pool, sizeof(struct pword *) * pword_count);
  pword_count = 0;
  do_for_all_pwords(pword_sdb);
  DBG("Second pass: %d words [should be same]", pword_count);
}

static byte *
pseudo_load(void)
{
  struct stat st;

  pseudo_pool = new_pool(16384);
  pseudo_fd = open(pseudo_list_name, O_RDONLY);
  if (pseudo_fd < 0)
    return "open failed";
  if (fstat(pseudo_fd, &st) < 0)
    return "stat failed";
  pseudos_size = st.st_size + 1;
  if ((pseudos = mmap(NULL, pseudos_size, PROT_READ, MAP_SHARED, pseudo_fd, 0)) == (void *) -1)
    return "mmap failed";
  DBG("Mapped pseudos at %p, size %d", pseudos, pseudos_size);
  pseudo_objs();
  pseudo_words();
  return NULL;
}

void
pseudo_rescan(int argc, byte **argv)
{
  byte *m;

  if (!pseudo_list_name)
    {
      add_qr("-110 No pseudo-object list defined, nothing to rescan");
      return;
    }
  pseudo_free();
  m = pseudo_load();
  if (!m)
    add_qr("+OK");
  else
    add_qr("-110 Error loading pseudo-objects: %s", m);
}

void
pseudo_init(void)
{
  byte *m;

  if (!pseudo_list_name)
    return;
  if (m = pseudo_load())
    log(L_ERROR "Error loading pseudo-objects: %s", m);
}

int
lookup_pseudo(byte *w, sh_off_t *refs, int j)
{
  struct pword *p;

  if (!num_pseudos)
    return j;
  p = pword_find(w, 0);
  if (!p)
    return j;
  DBG("<%s> matched ref %x", w, p->refnum);
  if ((uns) j >= max_word_matches)
    return -1;
  refs[j] = p->refnum;
  return j+1;
}

struct sdbitem *
lookup_pseudo_ref(sh_off_t ref0, uns *cnt)
{
  struct pword *p;
  uns ref = ref0 - FIRST_PSEUDO_REF;

  DBG("Resolving refs for %x", ref);
  if (ref >= pword_count)
    die("lookup_pseudo_ref: aargh");
  p = pword_index[ref];
  *cnt = p->refcnt;
  DBG("There are %d of them", *cnt);
  return p->refs;
}

int
show_pseudo_object(struct query *q, oid_t obj)
{
  byte *b, *e;

  DBG("Showing pseudo %x", obj);
  obj -= FIRST_PSEUDO_OBJ;
  if (obj >= num_pseudos)
    return 0;
  b = pseudo_ptrs[obj];
  e = pseudo_ptrs[obj+1];
  while (e > b + 1 && e[-1] == '\n' && e[-2] == '\n')
    e--;
  reply_string(q, b, e-b);
  return 1;
}
