/*
 *	Sherlock Search Engine -- Requests And Replies
 *
 *	(c) 1997 Martin Mares, <mj@atrey.karlin.mff.cuni.cz>
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <unistd.h>

#include "sherlock.h"

#include <lib/pools.h>

struct query *current_query;
struct replybuf *query_rbuf, *cache_rbuf;

void
init_reply_buf(struct replybuf *r, struct mempool *p)
{
  r->first = NULL;
  r->last = &r->first;
  r->pool = p;
}

struct query *
new_query(void)
{
  struct mempool *pool = new_pool(4096);
  struct query *q = pool_alloc(pool, sizeof(struct query));

  bzero(q, sizeof(struct query));
  q->pool = pool;
  init_reply_buf(&q->rbuf, pool);
  q->obptr = q->outbuf;
  q->obend = q->outbuf + BUF_SIZE;
  q->ibptr = q->inbuf;
  q->ibend = q->inbuf + BUF_SIZE;
  q->debug = global_debug;
  q->default_accent_mode = global_accent_mode;
  return q;
}

void
free_query(struct query *q)
{
  free_pool(q->pool);
}

static struct reply *
make_reply(struct replybuf *rb, byte *msg, va_list args)
{
  byte buf[4096];
  struct reply *r;
  uns l;

  l = vsprintf(buf, msg, args);
  if (log_requests > ((buf[0] == '-') ? 2U : 5U))
    log(L_INFO ">> %s", buf);
  r = fast_alloc(rb->pool, sizeof(struct reply) + l + 1);
  memcpy(r->text, buf, l);
  r->text[l++] = '\n';
  r->text[l] = 0;
  r->len = l;
  return r;
}

static void
add_reply_v(struct replybuf *rb, byte *msg, va_list args)
{
  struct reply *r;
  
  r = make_reply(rb, msg, args);
  *rb->last = r;
  rb->last = &r->next;
  r->next = NULL;
}

void
add_reply(struct replybuf *rb, byte *msg, ...)
{
  va_list args;

  va_start(args, msg);
  add_reply_v(rb, msg, args);
  va_end(args);
}

void
prepend_reply(struct replybuf *rb, byte *msg, ...)
{
  va_list args;
  struct reply *r;
  
  va_start(args, msg);
  r = make_reply(rb, msg, args);
  r->next = rb->first;
  if (!rb->first)
      rb->last = &r->next;
  rb->first = r;
  va_end(args);
}

static void
flush_buffer(struct query *q)
{
  uns w, l;

  if (q->fd < 0)			/* Marked as errorneous */
    return;
  l = q->obptr - q->outbuf;
  if (l)
    {
      w = write(q->fd, q->outbuf, l);
      if (l != w)			/* Write error... */
	q->fd = -1;
      q->obptr = q->outbuf;
    }
}

void
reply_string(struct query *q, byte *data, uns len)
{
  uns r;

  while (len)
    {
      r = q->obend - q->obptr;
      if (r > len)
	r = len;
      if (!r)
	{
	  flush_buffer(q);
	  continue;
	}
      memcpy(q->obptr, data, r);
      q->obptr += r;
      data += r;
      len -= r;
    }
}

void
ship_reply_buf(struct query *q, struct replybuf *rb)
{
  struct reply *r;

  for(r = rb->first; r; r=r->next)
    reply_string(q, r->text, r->len);
}

void
flush_reply_buf(struct query *q, struct replybuf *rb)
{
  ship_reply_buf(q, rb);
  init_reply_buf(rb, rb->pool);
}

void
flush_query(struct query *q)
{
  flush_buffer(q);
}

void
add_qr(byte *msg, ...)
{
  va_list args;

  va_start(args, msg);
  add_reply_v(query_rbuf, msg, args);
  va_end(args);
}

void
add_cr(byte *msg, ...)
{
  va_list args;

  va_start(args, msg);
  add_reply_v(cache_rbuf, msg, args);
  va_end(args);
}
