/*
 *	Sherlock Search Engine -- Declarations
 *
 *	(c) 1997--1999 Martin Mares, <mj@atrey.karlin.mff.cuni.cz>
 */

#include <lib/lib.h>
#include <lib/lists.h>

#ifdef BIG_DEBUG
#define DBG(x, y...) log(L_DEBUG x, ##y)
#else
#define DBG(x, y...) do { } while(0)
#endif

#ifdef DETAIL_DEBUG			/* Define locally in the file being debugged... */
#define DDBG(x, y...) DBG(x, ##y)
#else
#define DDBG(x, y...) do { } while(0)
#endif

/* config.c */

void read_config(void);
int check_access(u32);

extern byte *obj_pool_name, *obj_idx_name, *ref_list_name, *sig_tree_name;
extern byte *word_list_name, *nword_list_name, *nword_hash_name, *pseudo_list_name;
extern uns port, listen_queue, log_incoming, log_rejected, log_requests;
extern uns num_matches, cache_size, max_cache_age, max_words, max_word_matches;
extern uns file_cache_size, found_word_bonus, default_word_weight, log_timing;
extern uns connection_timeout, global_debug, global_accent_mode, max_wildcard_zone;
extern uns max_output_matches, allow_approx, send_all_words, min_wildcard_prefix_len;
extern uns ref_weight_scale;
extern byte *control_password;

/* request.c */

#define BUF_SIZE 4096

struct reply {
  struct reply *next;
  uns len;
  byte text[1];
};

struct replybuf {
  struct reply *first, **last;
  struct mempool *pool;
};

struct query {
  node n;
  uns established;			/* Time this connection was established */
  byte sections[256/8];			/* Section tag bitmap */
  int sections_set;			/* Section tag bitmap set */
  uns from, to;				/* Result range selection */
  uns debug;				/* Debug flags */
  struct expr *expr;			/* Query expression (NULL if it's a command) */
  char *cmd;				/* Command */
  struct mempool *pool;			/* Memory pool */
  struct replybuf rbuf;			/* Reply header */
  int fd;
  int default_accent_mode;		/* Default word accent mode */
  byte outbuf[BUF_SIZE];		/* Output buffering */
  byte *obptr, *obend;
  byte inbuf[BUF_SIZE];
  byte *ibptr, *ibend;
  uns parse_msec, map_msec,		/* Timing */
      words_msec, refs_msec, objs_msec;
};

extern struct query *current_query;

struct query *new_query(void);
void free_query(struct query *);
void init_reply_buf(struct replybuf *, struct mempool *);
void add_reply(struct replybuf *, byte *, ...);
void prepend_reply(struct replybuf *, byte *, ...);
void reply_string(struct query *, byte *, uns);
void ship_reply_buf(struct query *, struct replybuf *);
void flush_reply_buf(struct query *, struct replybuf *);
void flush_query(struct query *);

extern struct replybuf *query_rbuf, *cache_rbuf;
void add_qr(byte *, ...);
void add_cr(byte *, ...);

#define DFLAG_CACHE 1
#define DFLAG_WORDS 2

#ifdef DEBUG
#define DFLAG(flag) (current_query->debug & DFLAG_##flag)
#define DB(flag, x, y...) if (DFLAG(flag)) { log(L_DEBUG x, ##y); add_reply(query_rbuf, "." x, ##y); }
#else
#define DFLAG(flag) 0
#define DB(flag, x, y...)
#endif

/* lex.c */

void lex_init(byte *);
int yylex(void);

/* parse.y */

struct expr {
  byte type;
  union {
    struct {
      struct expr *l, *r;
    } op;
    struct {
      byte *word;
      word quality;
      byte classmap;
      byte is_nonword;
      byte wd_index;
      byte accent_mode;
    } match;
  } u;
};

#define EX_AND 0
#define EX_OR 1
#define EX_NOT 2
#define EX_MATCH 3
#define EX_ANY 4

#define ACCENT_AUTO 0			/* STRICT for accented query/match, STRIP elsewhere */
#define ACCENT_STRIP 1			/* strip all accents before comparing */
#define ACCENT_STRICT 2			/* compare with accents */

void err(byte *) NONRET;
byte *parse_query(byte *);

/* query.c */

#define HARD_MAX_WORDS 16		/* Need memory for 2^this bits */
#define HARD_MAX_WORD_MATCHES 256

void normalize_query(struct query *, byte *);
void cache_init(void);
void do_query(struct query *);
void ref_init(void);

/* words.c */

extern int have_word_list;

int lookup_word(byte *, sh_off_t *, uns);
void word_init(void);
byte *normalize_word(struct mempool *, byte *);

/* nonwords.c */

int lookup_nonword(byte *, sh_off_t *);
void nonword_init(void);
byte *normalize_nonword(struct mempool *, byte *);

/* files.c */

void file_cache_init(void);
int open_file(byte *);
byte *read_page(int, sh_off_t);
byte *read_data(int, sh_off_t, uns *);
void *lock_last_page(void);
void unlock_page(void *);

/* objects.c */

void objects_init(void);
void show_object(struct query *, oid_t);

/* cmds.c */

void do_command(struct query *);

/* pseudo.c */

#ifdef SHERLOCK_CONFIG_LARGE_DB
/* I hope this will go away one day... */
#define FIRST_PSEUDO_REF 0xff00000000000000ULL
#define IS_PSEUDO_REF(x) ((u64)(x) >= FIRST_PSEUDO_REF)
#else
#define FIRST_PSEUDO_REF 0xff000000
#define IS_PSEUDO_REF(x) ((u32)(x) >= FIRST_PSEUDO_REF)
#endif
#define FIRST_PSEUDO_OBJ 0xff000000

int lookup_pseudo(byte *, sh_off_t *, int);
struct sdbitem *lookup_pseudo_ref(sh_off_t, uns *);
int show_pseudo_object(struct query *, oid_t);
void pseudo_init(void);
void pseudo_rescan(int, byte **);
